#!/bin/bash

### SIDfm VM Update Manager ###

### Defines ###
CMD_NAME=$0
ARGS="$*"

# load profiles
[ -f /etc/profile.d/lang.sh ] && source /etc/profile.d/lang.sh
[ -f /etc/profile.d/railsenv.sh ] && source /etc/profile.d/railsenv.sh
[ -f /etc/profile.d/rh-ruby23.sh ] && source /etc/profile.d/rh-ruby23.sh
[ -f /etc/profile.d/proxy.sh ] && source /etc/profile.d/proxy.sh
[ -f /etc/profile.d/timezone.sh ] && source /etc/profile.d/timezone.sh

# Paths
BASE_PATH=/var/lib/redmine
PLUGIN_PATH=${BASE_PATH}/plugins/sidfm-vm-server
LOGFILE=/update.log
BACKUPFILE=/backup.tar.gz
DLURL=https://www.softek.co.jp/SID/support/sidfmvm/download/server
DEVELOP=/flag.develop

### functions ###
# logger
function log() {
    mes="$*"
    echo -n `date +'%Y/%m/%d %H:%M:%S'` >> $LOGFILE
    echo " [update_manage] ${mes}" >> $LOGFILE
}

# usage
function usage() {
    echo "$0 <function> <args>"
    exit 1
}

# versions
CURRENT_VERSION=0.0.0
LASTEST_VERSION=0.0.0
function get_current_version() {
    temp=$(< /sidfmvm-version)
    temp2=$( echo $temp | sed -r 's/^[0-9]+\.[0-9]+\.[0-9]+$//')
    if [ "x$temp2" == "x" ]; then
	CURRENT_VERSION=$temp
    fi
}
function get_lastest_version() {
    temp=$(curl -s https://www.softek.co.jp/SID/support/sidfmvm/download/server/sidfmvm-version)
    temp2=$( echo $temp | sed -r 's/^[0-9]+\.[0-9]+\.[0-9]+$//')
    if [ "x$temp2" == "x" ]; then
	LASTEST_VERSION=$temp
    fi
}
function get_versions() {
    get_current_version
    get_lastest_version
    echo "Current Version: $CURRENT_VERSION"
    echo "Lastest Version: $LASTEST_VERSION"
}
get_versions

# difference
DIFF_VERSIONS=
function diff_versions() {
    if [ ! -f "${DEVELOP}" -a "x$CURRENT_VERSION" != "x$LASTEST_VERSION" ]; then
	DIFF_VERSIONS=true
    fi
}
diff_versions

# subs
function diff_check() {
    if [ "x$DIFF_VERSIONS" == "x" ]; then
	    echo No need to update.
	    exit 0
    fi
}
redmine_filename=
redmine_url=
function prepare() {
    temp_path="${DLURL}/redmine-filename"
    redmine_filename=$(curl -s $temp_path)
    temp=$(echo "$redmine_filename" | sed -r 's/^redmine-([0-9a-z\.-]+)\.tar\.gz$//')
    if [ "x$temp" != "x" ]; then
	echo "Access Failure: $temp_path"
	exit 1
    fi
    temp_path="${DLURL}/redmine-url"
    redmine_url=$(curl -s $temp_path)
    temp=$(echo "$redmine_url" | sed -r 's/^https?:\/\/(.*).tar.gz$//')
    if [ "x$temp" != "x" ]; then
	echo "Access Failure: $temp_path"
	exit 1
    fi
}
dbconfig_path=/var/lib/redmine/config/database.yml
backup_dbconfig_path=/var/lib/database.yml
secrets_path=/var/lib/redmine/config/secrets.yml
backup_secrets_path=/var/lib/secrets.yml
#configuration_path=/var/lib/redmine/config/configuration.yml
#backup_configuration_path=/var/lib/configuration.yml
inflections_path=/var/lib/redmine/config/initializers/inflections.rb
backup_inflections_path=/var/lib/inflections.rb
function config_backup() {
    cp -f $dbconfig_path $backup_dbconfig_path
    cp -f $secrets_path $backup_secrets_path
    #cp -f $configuration_path $backup_configuration_path
    cp -f $inflections_path $backup_inflections_path
}
function config_restore() {
    cp -f $backup_dbconfig_path $dbconfig_path
    cp -f $backup_secrets_path $secrets_path
    #cp -f $backup_configuration_path $configuration_path
    cp -f $backup_inflections_path $inflections_path
}

function server_disable() {
    /vm_manage issue_update disable
    /vm_manage server disable
    echo wait 5 sec.
    sleep 5
}
function server_enable() {
    /vm_manage server enable
    /vm_manage issue_update enable
}

function redmine_unlink() {
    /bin/bash /var/lib/redmine/plugins/sidfm-vm-server/vminit.sh disable
}
function redmine_link() {
    /bin/bash /var/lib/redmine/plugins/sidfm-vm-server/vminit.sh enable
}

function change_owner() {
    chown -R apache:apache /var/lib/redmine
}

function cache_clear() {
    cd /var/lib/redmine
    bundle exec rails runner 'Rails.cache.clear'
}

function bundle_update() {
    cd /var/lib/redmine
    bundle update
}

function db_migrate() {
    cd /var/lib/redmine
    bundle exec rake db:migrate
    bundle exec rake redmine:plugins:migrate
}

function remove_plugin_assets() {
    rm -rf /var/lib/redmine/public/plugin_assets/sidfm
}

function redmine_update() {
    cd /var/lib
    redmine_name=$(echo `basename "$redmine_filename" .tar.gz`)
    curl -O "$redmine_url"
    tar xzf "$redmine_filename"
    cp -Rf "$redmine_name" redmine
    rm -rf $redmine_name $redmine_filename
}

function sidfmvm_update() {
    sidfmvm_filename=sidfmvm-${LASTEST_VERSION}.tar.gz
    cd /var/lib
    curl -O "${DLURL}/${sidfmvm_filename}"
    tar xzf "$sidfmvm_filename" -C /tmp/
    rm -rf /var/lib/redmine/plugins/sidfm-vm-server
    cd /var/lib/redmine
    tar xzf /tmp/sidfmvm/sidfmvm.tar.gz
    cp -fp /tmp/sidfmvm/vminit.sh /var/lib/redmine/plugins/sidfm-vm-server/
    chmod 755 /var/lib/redmine/plugins/sidfm-vm-server/vminit.sh
    rm -rf /tmp/sidfmvm "/var/lib/$sidfmvm_filename"
}

### main ###
ARG_FUNC=$1
if [ "x$1" == "x" ]; then
    usage
fi
shift

case "$ARG_FUNC" in
    "status")
	if [ "x$DIFF_VERSIONS" != "x" ]; then
	    echo There is an applicatable update.
	else
	    echo No need to update.
	fi
	;;
    "update")
	diff_check
	prepare
	server_disable
	cache_clear
	config_backup
	redmine_unlink
	redmine_update
	sidfmvm_update
	redmine_link
	config_restore
	change_owner
	bundle_update
	db_migrate
	remove_plugin_assets
	server_enable
	exit 0
	;;
    *)
	usage
	exit 1
	;;
esac
