#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'pp'
require 'time'
require 'digest'
require 'fileutils'

require 'vm_rule'

module SIDfm
  class VMDaemon

    # 各種操作に使用するアカウント。基本的には変更しないこと。
    ADMIN_USER = User.find_by_login('sidfmadm')

    ISSUE_STATUS_FIXED = IssueStatus.find(3).name
    ISSUE_STATUS_NOTCONFIRMED = IssueStatus.find(1).name

    NOTE_NO_LONGER_DETECTED = "脆弱性情報との照合の結果、脆弱性が検出されなくなったため、ステータスを#{ISSUE_STATUS_FIXED}に変更しました。"

    NOTE_CHANGE_PRODVER = "対象の脆弱性について、影響のあるプロダクトの情報が更新されたため、ステータスを#{ISSUE_STATUS_NOTCONFIRMED}に変更しました。お手数ですが脆弱性の内容を再度ご確認の上、対応をお願いします。"

    WIN_RISKS = {
      'Critical' => 4,
      'Important' => 3,
      'Moderate' => 2,
      'Low' => 1,
      nil => 4,
    }.freeze

    VM_TRACKER = Tracker.find_by_name('SIDfm脆弱性管理')
    VM_PROJECT = Project.find_by_name('脆弱性情報管理')

    DEFAULT_WIN_AREA = SidfmdArea.find_by_area_id(0)

    DEFAULT_PRIORITY = IssuePriority.find_by_position_name('default')

    LOGDIR = '/var/lib/redmine/log'
    LOG_INFO = 'vm_daemon-info.log'
    LOG_ERROR = 'vm_daemon-error.log'
    LOG_DEBUG = 'vm_daemon-debug.log'
    LOG_SIZE = 10 * 1024 * 1024

    DEBUG = FileTest.exist?('/flag.debug')

    private

    def initialize
      # vm_daemonの処理で新規登録・更新・脆弱性除外の情報を保持
      # prev_uls_content : 同期前の脆弱性一覧
      # modify_htmlids: 変更されるsidfmd_issue_idに対応するhtml_idのデータの配列
      # remove_htmlids: 影響無に変更されたhtmlidsの配列
      @rule_vulsync = {
        :prev_vuls_content => SidfmdIssue.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required),
        :modify_htmlids => [],
        :remove_htmlids => [],
      }
      # timelineで取得されるコンテンツを保存して、vm_ruleの処理で利用する
      @rule_contents = { :vulsync_content => {}, }
    end

    def logging(file, *s)
      if FileTest.exist?(file) and File.size(file) > LOG_SIZE then
        8.times do |i|
          file_n = file + ".#{8 - i}"
          file_old = file + ".#{8 - i + 1}"
          if FileTest.exist?(file_n) then
            FileUtils.mv(file_n, file_old, :force => true)
          end
        end
        if FileTest.exist?(file) then
          FileUtils.mv(file, file + ".0", :force => true)
        end
      end
      File.open(file, 'a') do |f|
        s.each do |i|
          f.puts(s.kind_of?(String) ? i : i.inspect)
        end
      end
    end

    def log_info(*s)
      logging("#{LOGDIR}/#{LOG_INFO}", *s)
    end

    def log_error(*s)
      logging("#{LOGDIR}/#{LOG_ERROR}", *s)
    end

    def log_debug(*s)
      logging("#{LOGDIR}/#{LOG_DEBUG}", *s) if DEBUG
    end

    # アップデートが必要ならtrueを返す。
    # nilを含む場合、どちらもnilの場合のみfalse、片方のみの場合は全てtrueとなる。
    def check_update(o_dt, u_dt)
      return false if (o_dt.nil? and u_dt.nil?)
      return true if o_dt.nil?
      return true if u_dt.nil?
      return (o_dt < u_dt)
    end

    # 指定したsidfmd_issueについて、ステータスを未対策に変更する。
    def change_status_not_confirmed(issue)
      issue.change_status_with_journal(ADMIN_USER.id,
                                       SidfmdIssue::STATUS_NOTCONFIRMED,
                                       NOTE_CHANGE_PRODVER)
    end

    # 指定したsidfmd_issueについて、ステータスを対策済に変更する。
    def change_status_fixed(issue)
      @rule_vulsync[:remove_htmlids] << issue[:html_id]
      issue.change_status_with_journal(ADMIN_USER.id,
                                       SidfmdIssue::STATUS_FIXED,
                                       NOTE_NO_LONGER_DETECTED)
    end

    public

    # 影響範囲の同期
    def areas_update
      areas = nil
      Sidfm.sidapi do |o|
        areas = o.areas
      end
      if areas and areas['areas'] then
        areas = areas['areas']
        areas.each do |area|
          areas_update_unit(area)
        end
      end
    end

    # areaの内容とSidfmdAreaの内容を照合して、
    # 必要があれば追加・更新。
    def areas_update_unit(area)
      SidfmdArea.transaction do
        local_area = SidfmdArea.find_by_area_id(area['id'])
        if local_area then
          if local_area.name != area['name'] then
            unless local_area.update_attribute(:name, area['name']) then
              raise ActiveRecord::Rollback
            end
          end
        else
          area['area_id'] = area['id']
          area.delete('id')
          local_area = SidfmdArea.create(area)
        end
      end
    end

    # 危険度の同期
    def risks_update
      risks = nil
      Sidfm.sidapi do |o|
        risks = o.risks
      end
      if risks and risks['risks'] then
        risks = risks['risks']
        risks.each do |risk|
          risks_update_unit(risk)
        end
      end
    end

    # riskの内容とSidfmdRiskの内容を照合して、
    # 必要があれば追加・更新
    def risks_update_unit(risk)
      SidfmdRisk.transaction do
        local_risk = SidfmdRisk.find_by_risk_id(risk['id'])
        if local_risk then
          if local_risk.name != risk['name'] then
            unless local_risk.update_attribute(:name, risk['name']) then
              raise ActiveRecord::Rollback
            end
          end
          if local_risk.name_en != risk['name_en'] then
            unless local_risk.update_attribute(:name_en, risk['name_en']) then
              raise ActiveRecord::Rollback
            end
          end
        else
          risk['risk_id'] = risk['id']
          risk.delete('id')
          local_risk = SidfmdRisk.create(risk)
        end
      end
    end

    # プロダクト、バージョン情報の更新
    def products_update(force = nil)
      data = nil
      now = Time.current
      Sidfm.sidapi do |o|
        begin_datetime = Sidfm.prodver_last_update ? Sidfm.prodver_last_update.to_datetime.to_s : nil
        end_datetime = now.to_datetime.to_s
        if force or begin_datetime.nil? then
          data = o.products2
        else
          data = o.products2('begin_datetime' => begin_datetime, 'end_datetime' => end_datetime)
        end
      end
      if data and data['products'] then
        products = data['products']
        Sidfm.transaction do
          products.each do |product|
            products_update_unit(product)
          end
          Sidfm.prodver_last_update = now
        end
      end
    end

    # productの内容とSidfmdProduct,SidfmdVersionの内容を照合して、
    # 必要があれば追加・更新 ※ 削除はここではできない
    def products_update_unit(product)
      ActiveRecord::Base.transaction(requires_new: true) do
        versions = product['versions']
        exist_product = SidfmdProduct.find_by_pid(product['id'])
        unless exist_product then
          product['pid'] = product['id']
          product.delete('id')
          product.delete('versions')
          exist_product = SidfmdProduct.create(product)
        end
        if versions then
          versions.each do |version|
            exist_version = SidfmdVersion.find_by_vid(version['id'])
            if exist_version then
              if exist_version['update_datetime'] and exist_version['update_datetime'] < version['update_datetime'] then
                version.delete('id')
                version.delete('vid')
                unless exist_version.update_attributes(version) then
                  raise ActiveRecord::Rollback
                end
              end
            else
              version['vid'] = version['id']
              version['sidfmd_product_id'] = exist_product.id
              version.delete('id')
              exist_version = SidfmdVersion.create(version)
            end
          end
        end
      end
    end

    # queueの情報からSIDAPIで脆弱性情報を取得し、保存する
    # 保存先：SidfmdQueue*
    # 終了後、queue.done が false -> true に変化。
    def issues_update(target_hosts)
      flg_do_cache_clear = false
      target_hosts.each do |host|
        queue = nil
        win_queue = nil
        log_debug(" start: #{host.hostname}")
        opts = {
          :product_ids => [],
          :product_version_ids => [],
          :packages => [],
          :begin_datetime => Time.at(0).to_datetime.to_s,
          :end_datetime => Time.current.to_datetime.to_s,
        }
        if host.refdate then
          begin
            opts[:begin_datetime] = DateTime.parse(host.refdate.to_s + ' 00:00:00 +09:00:00').to_s
          rescue Exception => e
            log_error(e.message)
            if e.backtrace then
              e.backtrace do |i|
                log_errror(i)
              end
            end
          end
        end
        if host.is_agent then
          log_debug(" agent exist")
          # agent 有り
          queue = host.get_last_queue
          if queue then
            log_debug(" queue exist")
            # queueあり
            win_queue = queue.sidfmd_queue_win
            issues_update_with_queue(host, queue, opts)
          else
            log_debug(" not queue exist")
            # queue無し
            issues_update_without_queue(host, opts)
          end
        else
          log_debug(" no agent exist")
          # agent無し
          issues_update_without_agent(host, opts)
          if opts[:product_ids].size > 0 or
              opts[:product_version_ids].size > 0 then
            # agent無しの場合のqueueはここで生成する。
            queue = SidfmdQueue.create(:sidfmd_host_id => host.id,
                                       :done => false,
                                       :checked => false)
          end
        end
        if win_queue then
          # AgentありのWindows用処理
          flg_do_cache_clear = issues_update_win(host, queue, win_queue)
          ActiveRecord::Base.transaction do
            host.update_attribute :last_checked_at, Time.current
            host.update_attribute :last_checked_hash, calc_hash(opts)
          end
        end
        if queue and (opts[:product_ids].size > 0 or
                      opts[:product_version_ids].size > 0 or
                      (!opts[:packages].nil? and opts[:packages].size > 0)) then
          issues_update_main(host, queue, opts)
          flg_do_cache_clear = true
        end
        ActiveRecord::Base.transaction do
          host.update_attribute :last_checked_at, Time.current
          host.update_attribute :last_checked_hash, calc_hash(opts)
        end
        log_debug(" end: #{host.hostname}")
      end
      if flg_do_cache_clear then
        Rails.cache.clear
      end
    end

    # windows agent用 issues_update
    # ここではAgentから受信したデータ(html_id < 0)のみを取り扱う。
    def issues_update_win(host, queue, win_queue)
      # html_id < 0 のSidfmdIssueのみを対象とした処理。
      # 登録済のissueとqueueのデータとの比較
      now_kbs = host.sidfmd_issues
        .where('sidfmd_issues.html_id < 0')
        .select(:html_id)
        .collect {|x| x.html_id }.uniq
      new_kbs = win_queue.sidfmd_queue_win_details.collect {|x| -x.kb }.uniq
      append_kbs = new_kbs - now_kbs
      removed_kbs = now_kbs - new_kbs
      SidfmdIssue.transaction do
        # issue追加
        append_kbs.each do |kb|
          # 既に登録されているかどうか確認
          exist_issue = host.sidfmd_issues.where(:html_id => kb)
          if exist_issue.count > 0 then
            exist_issue = exist_issue.first
            # 登録されていればステータス確認し、済・無なら未対策へ変更
            o_issue = exist_issue.issue
            if SidfmdIssue::STATUS_FINISHED.include?(o_issue.status_id) then
              change_status_not_confirmed(exist_issue)
            end
          else
            # 登録されていなければ新規作成。
            wqdetails = win_queue.sidfmd_queue_win_details.where(:kb => -kb)
            if wqdetails.count > 0 then
              wqdetail = wqdetails.order(:updated_at => :desc).limit(1).first
              new_issue_opts = {
                :tracker_id => VM_TRACKER.id,
                :project_id => VM_PROJECT.id,
                :subject => "[#{host.hostname}]" + wqdetail.title,
                :description => "新規脆弱性登録",
                :author_id => ADMIN_USER.id,
                :priority_id => DEFAULT_PRIORITY.id,
                :status_id => SidfmdIssue::STATUS_NOTCONFIRMED,
              }
              new_issue = Issue.create(new_issue_opts)
              if new_issue then
                new_sidfmd_issue_opts = {
                  :issue_id => new_issue.id,
                  :sidfmd_host_id => host.id,
                  :html_id => kb,
                  :title => wqdetail.title,
                  :risk_id => WIN_RISKS[wqdetail.severity],
                  :area_id => DEFAULT_WIN_AREA.area_id,
                }
                new_sidfmd_issue = SidfmdIssue.create(new_sidfmd_issue_opts)
                unless new_sidfmd_issue then
                  logger.error("cannot create sidfmd_issue: SidfmdIssue.create(#{new_sidfmd_issue_opts.inspect})")
                end
              else
                logger.error("cannot create issue: Issue.create(#{new_issue_opts.inspect})")
              end
            else
              logger.error("cannot create issue: lack queue_win_details.where(:kb => -#{kb})")
            end
          end
        end
        # 対策済への変更
        removed_kbs.each do |kb|
          # issueを取得
          exist_issue = host.sidfmd_issues.where(:html_id => kb)
          if exist_issue.count > 0 then
            exist_issue = exist_issue.order(:id => :desc).limit(1)
            if exist_issue then
              exist_issue = exist_issue.first
              # ステータス確認し、未・中なら対策済へ変更
              o_issue = exist_issue.issue
              if SidfmdIssue::STATUS_UNFINISHED.include?(o_issue.status_id) then
                change_status_fixed(exist_issue)
              end
            else
              logger.error("cannot get exist issue: exist_isue.order(:id => :desc).limit(1))")
            end
          else
            logger.error("cannot get exist issue: host(#{host.id}).sidfmd_issues.where(:html_id => #{kb})")
          end
        end
      end
      ret = (append_kbs.size > 0 or removed_kbs.size > 0)
      ret
    end

    # issues_update agent有りqueue有り
    def issues_update_with_queue(host, queue, opts)
      check_dist_pkgs(queue, opts)
      check_applications(queue, opts)
      issues_update_without_agent(host, opts)
      if host.refdate then
        begin
          opts[:begin_datetime] = DateTime.parse(host.refdate.to_s + ' 00:00:00 +09:00:00').to_s
        rescue Exception => e
          log_error(e.message)
          if e.backtrace then
            e.backtrace.each do |i|
              log_error(i)
            end
          end
        end
      else
        # epoch
        opts[:begin_datetime] = Time.at(0).to_datetime.to_s
      end
    end

    # issues_update agent有りqueue無し
    def issues_update_without_queue(host, opts)
      # no-op
    end

    # issues_update agent無し
    def issues_update_without_agent(host, opts)
      apps = host.sidfmd_host_apps
      if apps.count > 0 then
        apps.each do |app|
          if app.version_id and app.version_id > 0 then
            opts[:product_version_ids] << app.version_id
          elsif app.product_id and app.product_id > 0 then
            opts[:product_ids] << app.product_id
          end
        end
      end
    end

    # product_ids, product_version_ids, packages でHash値を生成する。
    def calc_hash(opts)
      pids = opts[:product_ids]
      vids = opts[:product_version_ids]
      pkgs = opts[:packages]
      pids = [] unless pids
      vids = [] unless vids
      pkgs = [] unless pkgs
      pids = pids.uniq.sort.join(',')
      vids = vids.uniq.sort.join(',')
      pkgs = pkgs.uniq.sort.join(',')
      [ Digest::SHA256.hexdigest(pids), Digest::SHA256.hexdigest(vids), Digest::SHA256.hexdigest(pkgs) ].join('-')
    end

    # optsの情報をSIDAPIで照合して脆弱性を取得し、SidfmdQueue*に記録
    def issues_update_main(host, queue, opts)
      log_debug(" issues_update_main")
      vuls = nil
      new_hash = calc_hash(opts)
      old_hash =  host.last_checked_hash
      flag_diff = (old_hash != new_hash)
      ### FIXME: 差分更新は強制停止中
      flag_diff = true
      if !flag_diff then
        last_checked_at = host.last_checked_at
        if last_checked_at then
          opts[:begin_datetime] = last_checked_at.to_datetime.to_s
        end
      end
      Sidfm.sidapi(host.hostname, host.host_apikey) do |o|
        vuls = o.products_vulns(product_ids: opts[:product_ids].uniq,
                                product_version_ids: opts[:product_version_ids].uniq,
                                packages: opts[:packages],
                                begin_datetime: opts[:begin_datetime],
                                end_datetime: opts[:end_datetime],
                                time_field: 'affect_product',
                                package_scan_period: 'all')
      end
      if vuls and vuls['vulns'] then
        log_debug(" process api response")
        vuls = vuls['vulns']
        ActiveRecord::Base.transaction do
          ActiveRecord::Base.transaction(:requires_new => true) do
            # 古いSidfmdQueue*があれば削除する
            old_vuls = queue.sidfmd_queue_vuls
            old_vuls.each do |old_vul|
              old_vul.sidfmd_queue_vul_affprods.delete_all
              old_vul.sidfmd_queue_vul_affvers.delete_all
              old_vul.sidfmd_queue_vul_affpkgs.delete_all
              old_vul.sidfmd_queue_vul_cves.delete_all
              old_vul.sidfmd_queue_vul_sris.delete_all
            end
            old_vuls.delete_all
            # SidfmdQueue*に照合した結果を保存
            vuls.each do |vul|
              issues_update_main_vul(queue, vul)
            end
            unless queue.update_attribute(:done, true) then
              raise ActiveRecord::Rollback
            end
          end
          ActiveRecord::Base.transaction(:requires_new => true) do
            # SidfmdQueue*に保存した脆弱性情報と、既に登録されている脆弱性
            # (SidfmdIssue*)とを照合して、適切な処理を行う
            issues = host.sidfmd_issues
            if queue then
              # 以下の行では動かないので注意。
              #queue_vuls = queue.sidfmd_queue_vuls
              queue_vuls = SidfmdQueueVul.where(:sidfmd_queue_id => queue.id)
              if queue_vuls and queue_vuls.count > 0 then
                queue_vuls.each do |queue_vul|
                  issues_update_main_queue_vul(host, issues, queue_vul)
                end
              end
              ### 既に登録済のissueが今回検出されなかった場合。
              # 差分更新の場合はskip
              if flag_diff then
                issues_update_main_collation(host, issues, queue_vuls)
              end
              unless queue.update_attribute(:checked, true) then
                raise ActiveRecord::Rollback
              end
            end
            host.update_attribute :last_checked_at, Time.current
            host.update_attribute :last_checked_hash, new_hash
          end
        end
      end
    end

    # vulの情報をSidfmdQueue*に記録する
    def issues_update_main_vul(queue, vul)
      # vulのデータをVMのデータベース構成に合わせる
      vul['html_id'] = vul['id']
      vul.delete('id')
      vul['sidfmd_queue_id'] = queue.id
      affprods = vul['affect_products']
      vul.delete('affect_products')
      affvers = vul['affect_product_versions']
      vul.delete('affect_product_versions')
      affpkgs = vul['affect_packages']
      vul.delete('affect_packages')
      cves = vul['cves']
      vul.delete('cves')
      sris = vul['sris']
      vul.delete('sris')
      vul.delete('create_datetime')
      vul.delete('update_datetime')
      vul.delete('removed_affect_products')
      vul.delete('removed_affect_product_versions')
      vul.delete('short_description') if !SidfmdQueueVul.column_names.include?('short_description')
      l_vul = SidfmdQueueVul.create(vul)
      if l_vul then
        if affprods and affprods.size > 0 then
          affprods.each do |affprod|
            p = affprod['product']
            affprod.delete('product')
            affprod.delete('id')
            affprod['sidfmd_queue_vul_id'] = l_vul.id
            affprod['product_id'] = p['id']
            affprod['name'] = p['name']
            SidfmdQueueVulAffprod.create(affprod)
          end
        end
        if affvers and affvers.size > 0 then
          affvers.each do |affver|
            v = affver['product_version']
            affver.delete('product_version')
            affver.delete('id')
            affver['sidfmd_queue_vul_id'] = l_vul.id
            affver['product_version_id'] = v['id']
            affver['product_id'] = v['product_id']
            affver['fullname'] = v['fullname']
            affver['version'] = v['version']
            SidfmdQueueVulAffver.create(affver)
          end
        end
        if affpkgs and affpkgs.size > 0 then
          affpkgs.each do |affpkg|
            affpkg['sidfmd_queue_vul_id'] = l_vul.id
            affpkg.delete('id')
            SidfmdQueueVulAffpkg.create(affpkg)
          end
        end
        if cves and cves.size > 0 then
          cves.each do |cve|
            opts = { 'sidfmd_queue_vul_id' => l_vul.id }
            opts['cve_id'] = cve['cve_id']
            cvss2 = cve['cvss2']
            if cvss2 then
              opts['c2_vector'] = cvss2['vector']
              opts['c2_base_score'] = cvss2['base_score']
              opts['c2_create_datetime'] = cvss2['create_datetime']
              opts['c2_update_datetime'] = cvss2['update_datetime']
              opts['c2_exploitable'] = cvss2['exploitable']
            end
            cvss3 = cve['cvss3']
            if cvss3 then
              opts['c3_vector'] = cvss3['vector']
              opts['c3_base_score'] = cvss3['base_score']
              opts['c3_create_datetime'] = cvss3['create_datetime']
              opts['c3_update_datetime'] = cvss3['update_datetime']
              opts['c3_exploitable'] = cvss3['exploitable']
            end
            SidfmdQueueVulCve.create(opts)
          end
        end
        if sris and sris.size > 0 then
          sris.each do |sri|
            sri['sidfmd_queue_vul_id'] = l_vul.id
            sri.delete('id')
            SidfmdQueueVulSri.create(sri)
          end
        end
      end
    end

    def issues_update_main_queue_vul(host, issues, queue_vul)
      log_debug("issues_update_main_queue_vul: #{host.hostname} #{queue_vul.html_id}")
      exist_issue = issues.where(:html_id => queue_vul.html_id)
      if exist_issue.count == 0 then
        log_debug(" not exist")
        ### 対象ホストに登録されていないhtml_id(新規登録)
        opts_o_issue = {
          :tracker_id => VM_TRACKER.id,
          :project_id => VM_PROJECT.id,
          :subject => "[#{host.hostname}]" + queue_vul.title,
          :description => "新規脆弱性登録",
          :author_id => ADMIN_USER.id,
          :priority_id => DEFAULT_PRIORITY.id,
          :status_id => SidfmdIssue::STATUS_NOTCONFIRMED,
        }
        o_issue = Issue.create(opts_o_issue)
        log_debug(" not exist - pass 1")
        opts_l_issue = JSON.parse(queue_vul.to_json)
        opts_l_issue.delete('id')
        opts_l_issue.delete('sidfmd_queue_id')
        opts_l_issue['issue_id'] = o_issue.id
        opts_l_issue['sidfmd_host_id'] = host.id
        l_issue = SidfmdIssue.create(opts_l_issue)
        queue_vul_cves = queue_vul.sidfmd_queue_vul_cves
        if queue_vul_cves.count > 0 then
          log_debug(" not exist - pass 2")
          queue_vul_cves.each do |o_cve|
            opts_o_cve = JSON.parse(o_cve.to_json)
            opts_o_cve.delete('id')
            opts_o_cve.delete('sidfmd_queue_vul_id')
            opts_o_cve['sidfmd_issue_id'] = l_issue.id
            SidfmdIssueCve.create(opts_o_cve)
          end
        end
        queue_vul_sris = queue_vul.sidfmd_queue_vul_sris
        if queue_vul_sris.count > 0 then
          log_debug(" not exist - pass 3")
          queue_vul_sris.each do |o_sri|
            opts_o_sri = JSON.parse(o_sri.to_json)
            opts_o_sri.delete('id')
            opts_o_sri.delete('sidfmd_queue_vul_id')
            opts_o_sri['sidfmd_issue_id'] = l_issue.id
            SidfmdIssueSri.create(opts_o_sri)
          end
        end
        queue_vul_affprods = queue_vul.sidfmd_queue_vul_affprods
        if queue_vul_affprods.count > 0 then
          log_debug(" not exist - pass 4")
          queue_vul_affprods.each do |o_affprod|
            opts_o_affprod = JSON.parse(o_affprod.to_json)
            opts_o_affprod.delete('id')
            opts_o_affprod.delete('sidfmd_queue_vul_id')
            opts_o_affprod['sidfmd_issue_id'] = l_issue.id
            SidfmdIssueAffprod.create(opts_o_affprod)
          end
        end
        queue_vul_affvers = queue_vul.sidfmd_queue_vul_affvers
        if queue_vul_affvers.count > 0 then 
          log_debug(" not exist - pass 5")
          queue_vul_affvers.each do |o_affver|
            opts_o_affver = JSON.parse(o_affver.to_json)
            opts_o_affver.delete('id')
            opts_o_affver.delete('sidfmd_queue_vul_id')
            opts_o_affver['sidfmd_issue_id'] = l_issue.id
            SidfmdIssueAffver.create(opts_o_affver)
          end
        end
        queue_vul_affpkgs = queue_vul.sidfmd_queue_vul_affpkgs
        if queue_vul_affpkgs.count > 0 then
          log_debug(" not exist - pass 6")
          queue_vul_affpkgs.each do |o_affpkg|
            opts_o_affpkg = JSON.parse(o_affpkg.to_json)
            opts_o_affpkg.delete('id')
            opts_o_affpkg.delete('sidfmd_queue_vul_id')
            opts_o_affpkg['sidfmd_issue_id'] = l_issue.id
            SidfmdIssueAffpkg.create(opts_o_affpkg)
          end
        end
        system_category = host.system_category
        if system_category then
          log_debug(" not exist - pass 7")
          sri = l_issue.get_sri
          if sri then
            position_name = 'lowest'
            case sri.risk_id
            when 4 then 'high2'
            when 3 then 'high3'
            when 2 then 'default'
            when 1 then 'lowest'
            end
            unless o_issue.update_attribute(:priority_id, IssuePriority.find_by_position_name(position_name).id) then
              raise ActiveRecord::Rollback
            end
          end
        end
      else
        log_debug(" exist")
        ### 対象ホストに登録済のhtml_id(必要なら更新)
        exist_issue = exist_issue.first
        o_issue = exist_issue.issue
        o_issue_updated_on = o_issue.updated_on
        # 全体のアップデート確認フラグ。
        update_flag = false
        temp_flag = false
        change_prodver_flag = false
        affect_packages_count = 0
        # exist_issueとissueの更新
        if exist_issue.title != queue_vul.title then
          log_debug(" exist - pass 1")
          update_flag = true
          log_info("vm_daemon/title change: from #{exist_issue.title} to #{queue_vul.title}")
          exist_issue.update_attribute :title, queue_vul.title
          o_issue.update_attribute :subject, "[#{host.hostname}]" + queue_vul.title
        end
        if exist_issue.risk_id != queue_vul.risk_id then
          log_debug(" exist - pass 2")
          update_flag = true
          log_info("vm_daemon/risk_id change: from #{exist_issue.risk_id} to #{queue_vul.risk_id}")
          exist_issue.update_attribute :risk_id, queue_vul.risk_id
        end
        if exist_issue.area_id != queue_vul.area_id then
          log_debug(" exist - pass 3")
          update_flag = true
          log_info("vm_daemon/area_id change: from #{exist_issue.area_id} to #{queue_vul.area_id}")
          exist_issue.update_attribute :area_id, queue_vul.area_id
        end
        if SidfmdQueueVul.column_names.include?('short_description') and exist_issue.short_description != queue_vul.short_description then
          log_debug(" exist - pass 4")
          update_flag = true
          log_info("vm_daemon/short_description change: from #{exist_issue.short_description} to #{queue_vul.short_description}")
          exist_issue.update_attribute :short_description, queue_vul.short_description
        end
        # affect_product_versions
        unless update_flag then
          log_debug(" exist - pass 5")
          temp_flag = false
          if exist_issue.sidfmd_issue_affvers.count != queue_vul.sidfmd_queue_vul_affvers.count then
            temp_flag = true
          end
          if !temp_flag then
            queue_vul.sidfmd_queue_vul_affvers.each do |queue_affver|
              if check_update(o_issue_updated_on, queue_affver.update_datetime) then
                temp_flag = true
                break
              end
            end
          end
          if temp_flag then
            update_flag = true
          end
        end
        if update_flag then
          log_debug(" exist - pass 5")
          log_info("vm_daemon/affvers change: from #{exist_issue.sidfmd_issue_affvers.to_json} to #{queue_vul.sidfmd_queue_vul_affvers.to_json}")
          exist_issue.sidfmd_issue_affvers.delete_all
          queue_vul.sidfmd_queue_vul_affvers.each do |queue_affver|
            temp_opts = JSON.parse(queue_affver.to_json)
            temp_opts.delete('id')
            temp_opts.delete('sidfmd_queue_vul_id')
            temp_opts['sidfmd_issue_id'] = exist_issue.id
            SidfmdIssueAffver.create(temp_opts)
          end
          change_prodver_flag = true
        end
        # affect_products
        unless update_flag then
          log_debug(" exist - pass 6")
          temp_flag = false
          if exist_issue.sidfmd_issue_affprods.count != queue_vul.sidfmd_queue_vul_affprods.count then
            temp_flag = true
          end
          if !temp_flag then
            log_debug(" exist - pass 7")
            queue_vul.sidfmd_queue_vul_affprods.each do |queue_affprod|
              if check_update(o_issue_updated_on, queue_affprod.update_datetime) then
                temp_flag = true
                break
              end
            end
          end
          if temp_flag then
            log_debug(" exist - pass 8")
            update_flag = true
          end
        end
        if update_flag then 
          log_debug(" exist - pass 9")
         log_info("vm_daemon/affprods change: from #{exist_issue.sidfmd_issue_affprods.to_json} to #{queue_vul.sidfmd_queue_vul_affprods.to_json}")
          exist_issue.sidfmd_issue_affprods.delete_all
          queue_vul.sidfmd_queue_vul_affprods.each do |queue_affprod|
            temp_opts = JSON.parse(queue_affprod.to_json)
            temp_opts.delete('id')
            temp_opts.delete('sidfmd_queue_vul_id')
            temp_opts['sidfmd_issue_id'] = exist_issue.id
            SidfmdIssueAffprod.create(temp_opts)
          end
          change_prodver_flag = true
        end
        # affect_packages
        unless update_flag then
          log_debug(" exist - pass 10")
          temp_flag = false
          if exist_issue.sidfmd_issue_affpkgs.count != queue_vul.sidfmd_queue_vul_affpkgs.count then
            temp_flag = true
          end
          if temp_flag then
            update_flag = true
          end
        end
        if update_flag then
          log_debug(" exist - pass 11")
          log_info("vm_daemon/affpkgs change: from #{exist_issue.sidfmd_issue_affpkgs.to_json} to #{queue_vul.sidfmd_queue_vul_affpkgs.to_json}")
          exist_issue.sidfmd_issue_affpkgs.delete_all
          queue_vul.sidfmd_queue_vul_affpkgs.each do |queue_affpkg|
            temp_opts = JSON.parse(queue_affpkg.to_json)
            temp_opts.delete('id')
            temp_opts.delete('sidfmd_queue_vul_id')
            temp_opts['sidfmd_issue_id'] = exist_issue.id
            SidfmdIssueAffpkg.create(temp_opts)
          end
          affect_packages_count = exist_issue.sidfmd_issue_affpkgs.count
        end
        # cves
        unless update_flag then
          log_debug(" exist - pass 12")
          temp_flag = false
          if exist_issue.sidfmd_issue_cves.count != queue_vul.sidfmd_queue_vul_cves.count then
            temp_flag = true
          end
          if !temp_flag then
            queue_vul.sidfmd_queue_vul_cves.each do |queue_cve|
              if check_update(o_issue_updated_on, queue_cve.c3_update_datetime) then
                temp_flag = true
                break
              end
              if check_update(o_issue_updated_on, queue_cve.c2_update_datetime) then
                temp_flag = true
                break
              end
            end
          end
          if temp_flag then
            log_debug(" exist - pass 13")
            update_flag = true
          end
        end
        if update_flag then
          log_debug(" exist - pass 14")
          log_info("vm_daemon/cve change: from #{exist_issue.sidfmd_issue_cves.to_json} to #{queue_vul.sidfmd_queue_vul_cves.to_json}")
          exist_issue.sidfmd_issue_cves.delete_all
          queue_vul.sidfmd_queue_vul_cves.each do |o_cve|
            temp_opts = JSON.parse(o_cve.to_json)
            temp_opts.delete('id')
            temp_opts.delete('sidfmd_queue_vul_id')
            temp_opts['sidfmd_issue_id'] = exist_issue.id
            SidfmdIssueCve.create(temp_opts)
          end
        end
        # sris
        if update_flag then
          log_debug(" exist - pass 15")
          log_info("vm_daemon/sri change: from #{exist_issue.sidfmd_issue_sris.to_json} to #{queue_vul.sidfmd_queue_vul_sris.to_json}")
          exist_issue.sidfmd_issue_sris.delete_all
          queue_vul.sidfmd_queue_vul_sris.each do |o_sri|
            temp_opts = JSON.parse(o_sri.to_json)
            temp_opts.delete('id')
            temp_opts.delete('sidfmd_queue_vul_id')
            temp_opts['sidfmd_issue_id'] = exist_issue.id
            SidfmdIssueSri.create(temp_opts)
          end
        end
        # 優先度の更新
        if o_issue.priority_id != DEFAULT_PRIORITY.id then
          log_debug(" exist - pass 16")
          system_category = host.system_category
          if system_category then
            sri = exist_issue.get_sri
            if sri then
              position_name = 'lowest'
              case sri.risk_id
              when 4 then 'high2'
              when 3 then 'high3'
              when 2 then 'default'
              when 1 then 'lowest'
              end
              unless o_issue.update_attribute(:priority_id, IssuePriority.find_by_position_name(position_name).id) then
                log_debug(" exist - pass 17")
                raise ActiveRecord::Rollback
              end
            end
          end
        end
        # 影響のあるプロダクト・バージョンが変更され、かつ、影響のあるパッケージは無い場合
        if change_prodver_flag and affect_packages_count == 0 and SidfmdIssue::STATUS_FINISHED.include?(o_issue.status_id) then
          log_debug(" exist - pass 18")
          log_info("vm_daemon/append journal")
          # ステータスを未対策に変更
          change_status_not_confirmed(exist_issue)
        end
        if update_flag then
          log_debug(" exist - pass 19")
          log_info("vm_daemon/update issues.updated_on")
          o_issue.update_attribute :updated_on, Time.current
          @rule_vulsync[:modify_htmlids] << exist_issue[:html_id]
        end
      end
    end

    # queue_vulsとissuesの照合
    # 既に登録済のissueが今回検出されなかった場合の処理。
    def issues_update_main_collation(host, issues, queue_vuls)
      # html_id > 0は、Windows Agentの部分を排除するため。Windows Agentの部分に関しては
      # 別実装となる。
      removed_issues = issues
        .joins(:issue)
        .where('sidfmd_issues.html_id > 0')
        .where.not(:html_id => queue_vuls.collect {|x| x.html_id }.uniq)
        .where(:issues => { :status_id => SidfmdIssue::STATUS_UNFINISHED })
      if removed_issues.count > 0 then
        removed_issues.each do |issue|
          # ステータスを対策済に変更
          change_status_fixed(issue)
        end
      end
    end

    # agentで取得したdistを解析し、opts[:product_ids, :product_version_ids]に
    # 反映する
    def check_dist_pkgs(queue, opts)
      dist = queue.sidfmd_queue_dist
      if dist and dist.ident then
        case dist.ident
        when  'centos' then
          sidfmd_product = SidfmdProduct.find_by_name('CentOS')
          if dist.version then
            ver_splited = dist.version.split(/\./)
            if ver_splited.size > 0 then
              sidfmd_version = sidfmd_product.sidfmd_versions.find_by_version(ver_splited.first)
              if sidfmd_version then
                opts[:product_version_ids] << sidfmd_version.vid
              else
                # CentOS 8 の場合、Red Hat Enterprise Linux 8 での解決を試みる。
                r_sidfmd_product = SidfmdProduct.find_by_name('Red Hat Linux')
                sidfmd_version = r_sidfmd_product.sidfmd_versions.find_by_version("Enterprise Linux #{ver_splited.first}")
                if sidfmd_version then
                  opts[:product_version_ids] << sidfmd_version.vid
                else
                  # Red Hat Enterprise Linux でも適合しなければ、バージョンの認識を断念しプロダクトで登録。
                  opts[:product_ids] << sidfmd_product.pid
                end
              end
            end
          end
          opts[:packages] = queue.sidfmd_queue_pkgs.select(:pkg).collect {|x| x.pkg }
        when 'amzn' then
          sidfmd_product = SidfmdProduct.find_by_name('Amazon Linux AMI')
          sidfmd_version = sidfmd_product.sidfmd_versions.find_by_version(dist.version)
          if sidfmd_version then
            opts[:product_version_ids] << sidfmd_version.vid
          else
            opts[:product_ids] << sidfmd_product.pid
          end
          opts[:packages] = queue.sidfmd_queue_pkgs.select(:pkg).collect {|x| x.pkg }
        when 'rhel' then
          sidfmd_product = SidfmdProduct.find_by_name('Red Hat Linux')
          if dist.version then
            ver_splited = dist.version.split(/\./)
            if ver_splited.size > 0 then
              # Enterprise Linux Server 6/7
              sidfmd_version = sidfmd_product.sidfmd_versions.find_by_version("Enterprise Linux Server #{ver_splited.first}")
              if sidfmd_version then
                opts[:product_version_ids] << sidfmd_version.vid
              else
                # Enterprise Linux 5/8
                sidfmd_version = sidfmd_product.sidfmd_versions.find_by_version("Enterprise Linux #{ver_splited.first}")
                if sidfmd_version then
                  opts[:product_version_ids] << sidfmd_version.vid
                else
                  opts[:product_ids] << sidfmd_product.pid
                end
              end
            end
          end
          opts[:packages] = queue.sidfmd_queue_pkgs.select(:pkg).collect {|x| x.pkg }
        when 'ubuntu' then
          sidfmd_product = SidfmdProduct.find_by_name('Ubuntu')
          sidfmd_version = sidfmd_product.sidfmd_versions.find_by_version(dist.version + ' LTS')
          if sidfmd_version then
            opts[:product_version_ids] << sidfmd_version.vid
          else
            opts[:product_ids] << sidfmd_product.pid
          end
          opts[:packages] = queue.sidfmd_queue_pkgs.select(:pkg).collect {|x| x.pkg }
        when 'debian' then
          sidfmd_product = SidfmdProduct.find_by_name('Debian GNU/Linux')
          sidfmd_version = sidfmd_product.sidfmd_versions.find_by_version("#{dist.version}.0")
          if sidfmd_version then
            opts[:product_version_ids] << sidfmd_version.vid
          else
            opts[:product_ids] << sidfmd_product.pid
          end
          opts[:packages] = queue.sidfmd_queue_pkgs.select(:pkg).collect {|x| x.pkg }
        end
      end
    end

    # agentで取得したアプリケーション情報を解析し、opts[:product_ids, :product_version_ids]に
    # 反映する
    def check_applications(queue, opts)
      apps = queue.sidfmd_queue_apps
      if apps and apps.count > 0 then
        apps.each do |app|
          case app.name
          when 'Apache Struts' then check_application_struts(app, opts)
          when 'Apache Tomcat' then check_application_tomcat(app, opts)
          when 'PHP' then check_application_php(app, opts)
          when 'ruby' then check_application_ruby(app, opts)
          when 'Oracle Java' then check_application_jdk(app, opts)
          #when 'Oracle Java SE JRE' then check_application_jre(app, opts)
          end
        end
      end
    end

    # Apache Struts
    def check_application_struts(app, opts)
      product = SidfmdProduct.find_by_name('Apache Struts')
      if product then
        version = product.sidfmd_versions.find_by_version(app.version)
        if version then
          opts[:product_version_ids] << version.vid
        else
          opts[:product_ids] << product.pid
        end
      end
    end

    # Apache Tomcat
    def check_application_tomcat(app, opts)
      product = SidfmdProduct.find_by_name('Tomcat')
      if product then
        version = product.sidfmd_versions.find_by_version(app.version)
        if version then
          opts[:product_version_ids] << version.vid
        else
          opts[:product_ids] << product.pid
        end
      end
    end

    # PHP
    def check_application_php(app, opts)
      product = SidfmdProduct.find_by_name('PHP')
      if product then
        version = product.sidfmd_versions.find_by_version(app.version)
        if version then
          opts[:product_version_ids] << version.vid
        else
          opts[:product_ids] << product.pid
        end
      end
    end

    # Ruby
    def check_application_ruby(app, opts)
      product = SidfmdProduct.find_by_name('Ruby')
      if product then
        version = product.sidfmd_versions.find_by_version(app.version)
        unless version then
          # 2.5.1p57 のようにパッチレベルがつくので、それを除去
          if app.version =~ /(\d+\.\d+\.\d+)/ then
            version = product.sidfmd_versions.find_by_version(Regexp.last_match(1))
          end
        end
        if version then
          opts[:product_version_ids] << version.vid
        else
          opts[:product_ids] << product.pid
        end
      end
    end

    # Oracke JDK
    def check_application_jdk(app, opts)
      product_jdk = SidfmdProduct.find_by_name('Oracle Java SE JDK')
      product_sdk = SidfmdProduct.find_by_name('Oracle Java SE SDK')
      version = nil
      if app.path =~ /jdk/i and app.path !~ /jre/i then 
        if product_jdk then
          version = product_jdk.sidfmd_versions.find_by_version(app.version)
        end
        unless version then
          if product_sdk then
            version = product_sdk.sidfmd_versions.find_by_version(app.version)
          end
        end
      end
      if version then
        opts[:product_version_ids] << version.vid
      else
        opts[:product_ids] << product_jdk.pid
      end
    end

    # Oracle JRE
    def check_application_jre(app, opts)
      product_jre = SidfmdProduct.find_by_name('Oracle Java SE JRE')
      product_sdk = SidfmdProduct.find_by_name('Oracle Java SE SDK')
      version = nil
      if app.path =~ /jre/ then
        if product_jdk then
          version = product_jdk.sidfmd_versions.find_by_version(app.version)
        end
        unless version then
          if product_sdk then
            version = product_sdk.sidfmd_versions.find_by_version(app.version)
          end
        end
      end
      if version then
        opts[:product_version_ids] << version.vid
      else
        opts[:product_ids] << product_jre.pid
      end
    end

    # ホスト一覧を取得し返す
    def get_hosts_content
      @rule_contents[:hosts_content] = SidfmdHost.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required) unless @rule_contents[:hosts_content]
      @rule_contents[:hosts_content]
    end

    # 脆弱性一覧を取得し返す
    def get_vuls_content
      @rule_contents[:vuls_content] = SidfmdIssue.create_all_content(no_cache: true, category_id: 0, role_admin: true, corresp_required: Sidfm.corresp_required) unless @rule_contents[:vuls_content]
      @rule_contents[:vuls_content]
    end

    # 脆弱性同期時に変更されたデータを取得し返す
    def get_vulsync_content
      get_vuls_content
      # add
      prev_htmlids = @rule_vulsync[:prev_vuls_content].collect {|x| x[:html_id] }
      @rule_contents[:vulsync_content][:add] =
        @rule_contents[:vuls_content].find_all {|x| !prev_htmlids.include?(x[:html_id]) }
      # remove
      @rule_contents[:vulsync_content][:remove] =
        @rule_contents[:vuls_content].find_all {|x|
        @rule_vulsync[:remove_htmlids].include?(x[:html_id])
      }
      # modify
      @rule_contents[:vulsync_content][:modify] = 
        @rule_contents[:vuls_content].find_all {|x|
        @rule_vulsync[:modify_htmlids].include?(x[:html_id])
      }
      @rule_contents[:vulsync_content]
    end

    # ホストのタイムライン処理
    def host_timeline_update
      hosts = get_hosts_content
      sri = { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0 }
      cvss = { :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :notreg => 0, :fixed => 0 }
      status = { :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0 }
      flg_fixed = (SidfmdHostTimeline.column_names.include?('sri_fixed') && SidfmdHostTimeline.column_names.include?('cvss_fixed'))
      hosts.each do |i|
        if flg_fixed then
          if SidfmdIssue::STATUS_UNFINISHED.include?(i[:min_status_id]) then
            case i[:max_sri]
            when 0 then sri[:none] += 1
            when 1 then sri[:low] += 1
            when 2 then sri[:medium] += 1
            when 3 then sri[:high] += 1
            when 4 then sri[:critical] += 1
            else sri[:none] += 1
            end
            if i[:max_cvss] then
              cvss[i[:max_cvss][:label]] += 1
            else
              cvss[:d1] += 1
            end
          else
            sri[:fixed] += 1
            cvss[:fixed] += 1
          end
        else
          case i[:max_sri]
          when 0 then sri[:none] += 1
          when 1 then sri[:low] += 1
          when 2 then sri[:medium] += 1
          when 3 then sri[:high] += 1
          when 4 then sri[:critical] += 1
          else sri[:none] += 1
          end
          if i[:max_cvss] then
            cvss[i[:max_cvss][:label]] += 1
          else
            cvss[:d1] += 1
          end
        end
        case i[:min_status_id]
        when SidfmdIssue::STATUS_NOTCONFIRMED then status[:notconfirmed] += 1
        when SidfmdIssue::STATUS_ONGOING then status[:ongoing] += 1
        when SidfmdIssue::STATUS_FIXED then status[:fixed] += 1
        when SidfmdIssue::STATUS_NOTAFFECT then status[:notaffect] += 1
        else status[:notaffect] += 1
        end
      end
      begin
        SidfmdHostTimeline.transaction do
          data = {
            :dt => Time.current,
            :sri_critical => sri[:critical],
            :sri_high => sri[:high],
            :sri_medium => sri[:medium],
            :sri_low => sri[:low],
            :sri_none => sri[:none],
            :cvss_d4 => cvss[:d4],
            :cvss_d3 => cvss[:d3],
            :cvss_d2 => cvss[:d2],
            :cvss_d1 => cvss[:d1],
            :cvss_notreg => cvss[:notreg],
            :status_notconfirmed => status[:notconfirmed],
            :status_ongoing => status[:ongoing],
            :status_fixed => status[:fixed],
            :status_notaffect => status[:notaffect],
          }
          if flg_fixed then
            data[:sri_fixed] = sri[:fixed]
            data[:cvss_fixed] = cvss[:fixed]
          end
          SidfmdHostTimeline.create(data)
        end
      rescue Exception => e
        log_error(e)
      end
    end

    def vul_timeline_update
      i_htmlids = get_vuls_content
      sri = { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0 }
      cvss = { :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :notreg => 0, :fixed => 0 }
      status = { :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0 }
      flg_fixed = (SidfmdVulTimeline.column_names.include?('sri_fixed') && SidfmdVulTimeline.column_names.include?('cvss_fixed'))
      i_htmlids.each do |i|
        if flg_fixed then
          if SidfmdIssue::STATUS_UNFINISHED.include?(i[:min_status_id]) then
            case i[:max_sri]
            when 0 then sri[:none] += 1
            when 1 then sri[:low] += 1
            when 2 then sri[:medium] += 1
            when 3 then sri[:high] += 1
            when 4 then sri[:critical] += 1
            else sri[:none] += 1
            end
            if i[:cvss] then
              cvss[i[:cvss][:label]] += 1
            else
              cvss[:d1] += 1
            end
          else
            sri[:fixed] += 1
            cvss[:fixed] += 1
          end
        else
          case i[:max_sri]
          when 0 then sri[:none] += 1
          when 1 then sri[:low] += 1
          when 2 then sri[:medium] += 1
          when 3 then sri[:high] += 1
          when 4 then sri[:critical] += 1
          else sri[:none] += 1
          end
          if i[:cvss] then
            cvss[i[:cvss][:label]] += 1
          else
            cvss[:d1] += 1
          end
        end
        case i[:min_status_id]
        when SidfmdIssue::STATUS_NOTCONFIRMED then status[:notconfirmed] += 1
        when SidfmdIssue::STATUS_ONGOING then status[:ongoing] += 1
        when SidfmdIssue::STATUS_FIXED then status[:fixed] += 1
        when SidfmdIssue::STATUS_NOTAFFECT then status[:notaffect] += 1
        else status[:notaffect] += 1
        end
      end
      begin
        SidfmdVulTimeline.transaction do
          data = {
            :dt => Time.current,
            :sri_critical => sri[:critical],
            :sri_high => sri[:high],
            :sri_medium => sri[:medium],
            :sri_low => sri[:low],
            :sri_none => sri[:none],
            :cvss_d4 => cvss[:d4],
            :cvss_d3 => cvss[:d3],
            :cvss_d2 => cvss[:d2],
            :cvss_d1 => cvss[:d1],
            :cvss_notreg => cvss[:notreg],
            :status_notconfirmed => status[:notconfirmed],
            :status_ongoing => status[:ongoing],
            :status_fixed => status[:fixed],
            :status_notaffect => status[:notaffect],
          }
          if flg_fixed then
            data[:sri_fixed] = sri[:fixed]
            data[:cvss_fixed] = cvss[:fixed]
          end
          SidfmdVulTimeline.create(data)
        end
      rescue Exception => e
        log_error(e)
      end
    end

    # VMRuleの処理
    def vm_rules_invoke
      log_debug("vm_rule_invoke:")
      log_debug("modify_htmlids:", @rule_vulsync[:modify_htmlids])
      log_debug("remove_htmlids:", @rule_vulsync[:remove_htmlids])
      # periodic
      periodic_rule_args = {}
      periodic_rule_ids = SIDfm::VMRule.invoke_rule_ids(:periodic, periodic_rule_args)
      if periodic_rule_ids.size > 0 then
        log_info("periodic_rule_invoke:", periodic_rule_ids)
        get_hosts_content
        get_vuls_content
        periodic_rule_args[:hosts_content] = @rule_contents[:hosts_content]
        periodic_rule_args[:vuls_content] = @rule_contents[:vuls_content]
        SIDfm::VMRule.invoke_rule_run(periodic_rule_ids, periodic_rule_args)
        log_info("periodic_rule_invoke: end")
      end
      # vulsync
      vulsync_rule_args = {}
      vulsync_rule_ids = SIDfm::VMRule.invoke_rule_ids(:vulsync, vulsync_rule_args)
      if vulsync_rule_ids.size > 0 then
        log_info("vulsync_rule_invoke:", vulsync_rule_ids)
        get_hosts_content
        get_vuls_content
        get_vulsync_content
        log_debug("content_add(size):", @rule_contents[:vulsync_content][:add].size)
        log_debug("content_modify(size):", @rule_contents[:vulsync_content][:modify].size)
        log_debug("content_remove(size):", @rule_contents[:vulsync_content][:remove].size)
        vulsync_rule_args[:prev_vuls_content] = @rule_vulsync[:prev_vuls_content]
        vulsync_rule_args[:content_add] = @rule_contents[:vulsync_content][:add]
        vulsync_rule_args[:content_modify] = @rule_contents[:vulsync_content][:modify]
        vulsync_rule_args[:content_remove] = @rule_contents[:vulsync_content][:remove]
        SIDfm::VMRule.invoke_rule_run(vulsync_rule_ids, vulsync_rule_args)
        log_info("vulsync_rule_invoke: end")
      end
    end

    def execute
      target_hosts =
        SidfmdHost
        .enable_hosts
        .where('last_checked_at IS NULL OR last_checked_at < ?', Time.current.ago(1.day))
        .order(:last_checked_at)
        .limit(20)
      log_debug("vm_daemon: target=#{target_hosts.collect {|x| x.id.to_s + '-' + x.hostname }.join(', ')}")
      if target_hosts and target_hosts.count > 0 then
        areas_update
        risks_update
        products_update
        issues_update(target_hosts)
      end
      last_host_timeline = SidfmdHostTimeline.select(:dt).order(:dt => :desc).limit(1).first
      if !last_host_timeline or Time.current - last_host_timeline.dt > 24.0 * 60.0 * 60.0 then
        host_timeline_update
      end
      last_vul_timeline = SidfmdVulTimeline.select(:dt).order(:dt => :desc).limit(1).first
      if !last_vul_timeline or Time.current - last_vul_timeline.dt > 24.0 * 60.0 * 60.0 then
        vul_timeline_update
      end
      vm_rules_invoke
    end

  end
end
