# -*- coding: utf-8 -*-
require 'date'
require 'sidfm_controller_support'
require 'gen_ansible_playbook'

class SidfmHostController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private

  include SidfmControllerSupport
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license
  before_action :check_roles

  before_action :get_id

  before_action :get_view_state
  after_action :put_view_state

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    if [ 'update_status', 'update_due_dates' ].include?(action_name) and !@sidfm_roles.include?(:use) then
      redirect_to sidfm_vulnerabilities_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).order(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id] }
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
                     end
  end

  # idを取得する。また、アクセスしているユーザがidに対応するホストを参照可能か確認する。
  def get_id
    @id = params[:id].to_i
    if !@role_admin and !@user_host_ids.include?(@id) then
      logger.error("Access Denied: user_id=#{User.current.id}, target_host_id=#{@id}, user_granted_host_ids=#{@user_host_ids.join(',')}")
      redirect_to sidfm_hosts_path, :notice => af_log_user('対象ホストに対するアクセスを許可されていません。') and return
    end
    hosts = SidfmdHost.enable_hosts.categorized.where(:id => @id)
    @host = hosts.count > 0 ? hosts.first : nil
    redirect_to sidfm_hosts_path unless @host
  end

  # 対象となるコンテンツ全てを含む配列を返す。
  # カテゴリ選択のみ、反映させる
  def create_all_content
    @vuls[:all_content] = @host.create_content(:corresp_required => @corresp_required)
  end

  FILTER_KEYS = {}

  def init_filter_keys
    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:warning] = { :values => [ 'warning', 'others', 'all' ].map(&:freeze).freeze, :default => 'all'.freeze }
      FILTER_KEYS[:status] = {
        :values => (SidfmdIssue::STATUS_ALL.collect {|x| x.to_s } +
                    [ SidfmdIssue::STATUS_UNFINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ SidfmdIssue::STATUS_FINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ '0' ]).map(&:freeze).freeze, :default => '0'.freeze }
      FILTER_KEYS[:sri] = { :values => [ '1', '2', '3', '4', '0', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:cvss] = { :values => [ 'd1', 'd2', 'd3', 'd4', 'notreg', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:area] = { :values => [ '3', '2', '1', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:exploitable] = { :values => [ 'true', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:journal] = { :values => [ 'exist', '1day', '1week', '1month', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:due_date] = { :values => nil, :default => '::' }
      FILTER_KEYS[:keyword] = { :values => nil, :default => '' }
      FILTER_KEYS[:date] = { :values => nil, :default => '::' }
    end
  end

  def check_filters
    init_filter_keys
    f_keys = FILTER_KEYS.keys
    for i in 0..f_keys.size - 1 do
      if @vuls[:filters].has_key?(f_keys[i]) then
        if FILTER_KEYS[f_keys[i]][:values] && !FILTER_KEYS[f_keys[i]][:values].include?(@vuls[:filters][f_keys[i]]) then
          @vuls[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
        end
      else
        @vuls[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
      end
    end
  end

  DAY_SECONDS = 24 * 60 * 60

  # フィルタの適用
  def apply_filters
    ret = @vuls[:all_content].find_all {|item|
      # 要注意
      if @vuls[:filters][:warning] == 'warning' then
        next false unless item[:warning][@vuls[:idx]]
      elsif @vuls[:filters][:warning] == 'others' then
        next false if item[:warning][@vuls[:idx]]
      end
      # 対応状況
      if /\A(\d):(\d)\z/ =~ @vuls[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (item[:status_id] == status1 or item[:status_id] == status2)
      elsif @vuls[:filters][:status] != '0' then
        next false unless item[:status_id] == @vuls[:filters][:status].to_i
      end
      # SRI
      if @vuls[:filters][:sri] != '' then
        next false unless item[:sri] == @vuls[:filters][:sri].to_i
      end
      # CVSS
      if @vuls[:filters][:cvss] != '' then
        next false unless item[:cvss][:label] == @vuls[:filters][:cvss].to_sym
      end
      # SofTek影響範囲
      if @vuls[:filters][:area] != '' then
        next false unless (item[:html_id] > 0 and  @vuls[:filters][:area].to_i == item[:softek][:area_id])
      end
      # 攻撃コード
      if @vuls[:filters][:exploitable] != '' then
        next false unless item[:exploitable]
      end
      # 履歴
      if @vuls[:filters][:journal] != '' then
        if @vuls[:filters][:journal] == 'exist' then
          next false unless item[:journals].size > 0
        else
          temp_timespan = 1.days
          case @vuls[:filters][:journal]
          when '1day'   then temp_timespan = 1.days
          when '1week'  then temp_timespan = 1.weeks
          when '1month' then temp_timespan = 1.months
          end
          timespan = Time.current.ago(temp_timespan)
          next false unless item[:journals].find {|x|
            x[:created_on] >= timespan
          }
        end
      end
      # 対策予定日設定済
      if @vuls[:filters][:due_date] != '::' then
        if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @vuls[:filters][:due_date] then
          notset = ('notset' == Regexp.last_match(1))
          begindate = Regexp.last_match(2)
          enddate = Regexp.last_match(3)
          if notset then
            next false unless item[:due_date].nil?
          else
            next false if item[:due_date].nil?
            begin
              begindate = Date.parse(begindate)
            rescue Exception => e
              begindate = nil
            end
            begin
              enddate = Date.parse(enddate)
            rescue Exception => e
              enddate = nil
            end
            if begindate and enddate then
              next false unless (item[:due_date] >= begindate and item[:due_date] <= enddate)
            elsif begindate then
              next false unless item[:due_date] >= begindate
            elsif enddate then
              next false unless item[:due_date] <= enddate
            end
          end
        end
      end
      # キーワード検索
      if @vuls[:filters][:keyword] != '' then
        if /\A\s*(\d+)\s*\z/ =~ @vuls[:filters][:keyword] then
          html_id = Regexp.last_match(1).to_i
          next false unless (Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE) =~ item[:title] or html_id == item[:html_id])
        elsif /\A\s*KB(\d+)\s*\z/i =~ @vuls[:filters][:keyword] then
          kbid = -Regexp.last_match(1).to_i
          next false unless (Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE) =~ item[:title] or kbid == item[:html_id])
        else
          next false unless Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE) =~ item[:title]
        end
      end
      # 日付指定検索
      if @vuls[:filters][:date] != '::' then
        if /\A(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)\z/ =~ @vuls[:filters][:date] then
          orig_flg_rev = Regexp.last_match(1)
          orig_begin_date = Regexp.last_match(2)
          orig_end_date = Regexp.last_match(3)
          if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
            @vuls[:filters][:date] = '::'
            next false
          end
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if begin_date <= item[:created_on][:value] and end_date >= item[:created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false if begin_date <= item[:created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if end_date >= item[:created_on][:value]
            else
              # last resort
              @vuls[:filters][:date] = '::'
              next false
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless begin_date <= item[:created_on][:value] and end_date >= item[:created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false unless begin_date <= item[:created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless end_date >= item[:created_on][:value]
            else
              # last resort
              @vuls[:filters][:date] = '::'
              next false
            end
          end
        else
          @vuls[:filters][:date] = '::'
          next false
        end
      end
      true
    }
    #ret.sort! {|x, y|
    #  SidfmdIssue::SORT_PRIORITIES[x[:status_id]] <=> SidfmdIssue::SORT_PRIORITIES[y[:status_id]] ||
    #  y[:created_on][:value] <=> x[:created_on][:value]
    #}
    case @vuls[:order]
    when 'warning_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:warning][@vuls[:idx]] ? 500 : 1000 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'warning_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:warning][@vuls[:idx]] ? 500 : 1000 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'status_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:status_id] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'status_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:status_id] ]
      } 
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'sri_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:sri] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'sri_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:sri] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'cvss_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:cvss][:score] ? x[:cvss][:score] : -10.0  ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'cvss_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:cvss][:score] ? x[:cvss][:score] : -10.0  ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'area_asc' then
      temp_ev = proc {|x|
        [
         x[:title],
         x[:created_on][:display],
         x[:html_id] > 0 ? x[:softek][:area_id] : -10,
        ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'area_desc' then
      temp_ev = proc {|x|
        [
         x[:title],
         x[:created_on][:display],
         x[:html_id] > 0 ? x[:softek][:area_id] : -10,
        ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'exploitable_asc' then
      temp_ev = proc {|x|
        [
         x[:title], x[:created_on][:display],
         x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 1000) : 1500) : 1250
        ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'exploitable_desc' then
      temp_ev = proc {|x|
        [ x[:title],
          x[:created_on][:display],
          x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 1000) : 1500) : 1250 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'elapseddays_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:elapsed_days] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'elapseddays_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:elapsed_days] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'title_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:html_id] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'title_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:html_id] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'journals_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:journals].size ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'journals_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:created_on][:display], x[:journals].size ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'duedate_asc' then
      temp_ev = proc {|x|
        [ x[:title],
          x[:created_on][:display],
          x[:due_date] ? x[:due_date] : Date.new(1970,1,1),
          x[:due_date] ? 500 : 1000 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = xx[2] <=> yy[2]
          if r == 0 then
            r = yy[1] <=> yy[1]
            if r == 0 then
              r = xx[0] <=> yy[0]
            end
          end
        end
        r
      }
    when 'duedate_desc' then
      temp_ev = proc {|x|
        [ x[:title],
          x[:created_on][:display],
          x[:due_date] ? x[:due_date] : Date.new(1970,1,1),
          x[:due_date] ? 500 : 1000 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = xx[2] <=> yy[2]
          if r == 0 then
            r = yy[1] <=> yy[1]
            if r == 0 then
              r = xx[0] <=> yy[0]
            end
          end
        end
        r
      }
    end
    ret
  end

  TIMELINE_SPAN = 6
  TIMELINE_SPAN_UNIT = :month

  # タイムラインのデータを返す
  def get_timeline_data
    ret = {}
    ret[:now] = Time.current
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @vuls[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical', 5 => 'fixed',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大', 5 => '済',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1, :fixed ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
        :fixed => '済',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end
    @vuls[:all_content].each do |item|
      ret[:timescale].each do |step|
        next if item[:created_on][:value] > step[:time]
        # issue
        issue_status_id = item[:status_id]
        # journals
        temp_j = item[:journals].collect {|x|
          if x[:created_on] <= step[:time] and x[:details] and
              x[:details].find {|x| x[:prop_key] == 'status_id' } then
            x
          else
            nil
          end
        }
        temp_j.compact!
        temp_j.sort! {|x, y|
          x[:created_on] == y[:created_on] ?
          x[:id] <=> y[:id] :
          x[:created_on] <=> y[:created_on] }
        journal_status_id = nil
        if temp_j.size > 0 then
          temp_j.last[:details].each do |i|
            if i[:prop_key] == 'status_id' then
              journal_status_id = i[:value].to_i
              break
            end
          end
        end
        count_index = TIMELINE_SPAN - 1 - step[:step]
        if journal_status_id then
          ret[:status_counts][journal_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(journal_status_id) then
            case @vuls[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          else
            case @vuls[:idx]
            when :sri then
              ret[:risk_counts][5][count_index] += 1
            when :cvss then
              ret[:risk_counts][:fixed][count_index] += 1
            end
          end
        elsif issue_status_id then
          ret[:status_counts][issue_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(issue_status_id) then
            case @vuls[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          else
            case @vuls[:idx]
            when :sri then
              ret[:risk_counts][5][count_index] += 1
            when :cvss then
              ret[:risk_counts][:fixed][count_index] += 1
            end
          end
        end
      end
    end

    ret
  end

  # パラメータ解析
  def parse_params
    pages_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :grp, :idx, :filters, :checked, :all_checked)

    @vuls = { :url => request.url }

    @vuls[:grp] = (pages_params[:grp] || 'count').to_sym
    @vuls[:idx] = (pages_params[:idx] || @view_state[:idx].to_s).to_sym
    @view_state[:idx] = @vuls[:idx]
    @vuls[:offset] = (pages_params[:offset] || 0).to_i
    @vuls[:per_page] = (pages_params[:per_page] || 10).to_i
    @vuls[:order] = pages_params[:order] || 'title_asc'
    @vuls[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @vuls[:all_checked] = pages_params[:all_checked] || false
    @vuls[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
    check_filters

    @corresp_required = Sidfm.corresp_required

    create_all_content
    @vuls[:corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:warning_orig][@vuls[:idx]] ? 1 : 0 }
    @vuls[:fixed_corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:fixed_warning][@vuls[:idx]]  ? 1 : 0 }
    @vuls[:not_fixed_corresp_required_vul_size] =
      @vuls[:corresp_required_vul_size] - @vuls[:fixed_corresp_required_vul_size]
    @vuls[:all_vul_size] = @vuls[:all_content].size
  end

  def filter_content
    @status = {
      :exploit => [ { :value => @vuls[:not_fixed_corresp_required_vul_size] } ]
    }
    @vuls[:distribution_status] = {
      :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0,
    }
    @vuls[:distribution_risk] =
      case @vuls[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end
    @vuls[:all_content].each do |i|
      # status
      case i[:status_id]
      when SidfmdIssue::STATUS_NOTCONFIRMED then
        @vuls[:distribution_status][:notconfirmed] += 1
      when SidfmdIssue::STATUS_ONGOING then
        @vuls[:distribution_status][:ongoing] += 1
      when SidfmdIssue::STATUS_FIXED then
        @vuls[:distribution_status][:fixed] += 1
      when SidfmdIssue::STATUS_NOTAFFECT then
        @vuls[:distribution_status][:notaffect] += 1
      end
      # risk
      if SidfmdIssue::STATUS_UNFINISHED.include?(i[:status_id]) then
        case @vuls[:idx]
        when :sri then
          case i[:sri]
          when SidfmdIssueSri::VALUES[0] then
            @vuls[:distribution_risk][:critical] += 1
          when SidfmdIssueSri::VALUES[1] then
            @vuls[:distribution_risk][:high] += 1
          when SidfmdIssueSri::VALUES[2] then
            @vuls[:distribution_risk][:medium] += 1
          when SidfmdIssueSri::VALUES[3] then
            @vuls[:distribution_risk][:low] += 1
          when SidfmdIssueSri::VALUES[4] then
            @vuls[:distribution_risk][:none] += 1
          end
        when :cvss then
          case i[:cvss][:label]
          when :notreg then
            @vuls[:distribution_risk][:notreg] += 1
          when :d4 then
            @vuls[:distribution_risk][:d4] += 1
          when :d3 then
            @vuls[:distribution_risk][:d3] += 1
          when :d2 then
            @vuls[:distribution_risk][:d2] += 1
          when :d1 then
            @vuls[:distribution_risk][:d1] += 1
          end
        end
      else
        @vuls[:distribution_risk][:fixed] += 1
      end
    end

    @vuls[:content_filtered] = apply_filters
    @vuls[:total_size] = @vuls[:content_filtered].size

    @vuls[:statuses] = SidfmdIssue.statuses

    @due_dates = @vuls[:all_content].collect {|x| x[:due_date]}.compact
    @system_categories = SidfmdCategory.system
    @not_system_categories = SidfmdCategory.not_system
    if !@role_admin then
      @not_system_categories = @not_system_categories.where(:id => @user_category_ids)
    end
    @statuses = IssueStatus.all.order(:id)
    @timeline_data = (@vuls[:grp] == :distribution) ? get_timeline_data : []
  end

  def create_host_info
    ret = {}
    queues = @host.sidfmd_queues.order(:created_at => :desc)
    queue = queues.first
    if queue then
      ret[:queue] = queue
      ret[:os] = @host.get_os
      ret[:pkgs] = queue.sidfmd_queue_pkgs
      ret.delete(:pkgs) if ret[:pkgs].count == 0
      ret[:apps] = queue.sidfmd_queue_apps
      ret.delete(:apps) if ret[:apps].count == 0
    end
    apps = @host.sidfmd_host_apps
    if apps then
      ret[:lapps] = apps
      ret.delete(:lapps) if ret[:lapps].count == 0
    end
    ret
  end

  public

  def index
    f_params = params.fetch(:form, {}).permit(:history, :is_update_one_row) 
    parse_params
    filter_content
    if session['hostid_with_htmlid'] then
      if session['hostid_with_htmlid']['vul'] then
        @vuls[:content_filtered].each do |vul|
          h_id = vul[:html_id].to_s
          if session['hostid_with_htmlid']['vul'][h_id] then
            session['hostid_with_htmlid']['vul'][h_id][1] = true
          end
        end
      end
      if session['hostid_with_htmlid']['host'] && !f_params[:is_update_one_row] then
        session['hostid_with_htmlid']['host'].each_key do |hostid|
          session['hostid_with_htmlid']['host'][hostid][1] = false
        end
      end
    end
    @hostinfo = create_host_info
    @self_user_categories = SidfmdCategory.where(:id => (SidfmdCategoriesUser.where(:user_id => User.current).map(&:sidfmd_category_id))).map{|c| c.is_system ? Hash["system_c", c.name] : Hash["custom_c", c.name]}
    @hostinfo_for_keywords = {
      osname: @hostinfo[:os] ? @hostinfo[:os][:osname] : '---',
      synchro: (@host.last_checked_at && Time.current - @host.last_checked_at <= 2.0 * 24.0 * 60.0 * 60.0) ? '同期中' : '非同期',
      last_checked_at: @host.last_checked_at ? @host.last_checked_at.strftime('%Y/%m/%d %H:%M:%S') :  @hostinfo[:queue] ? @hostinfo[:queue].created_at.strftime('%Y/%m/%d %H:%M:%S') : '未同期',
      agent: @host.is_agent ? '使用' : '不使用',
      s_category: @host.has_system_category? ? @host.system_category.name : '---',
      c_category: @host.no_system_categories ? @host.no_system_categories.map{|c| c.sidfmd_category.name} : ['カスタムカテゴリ無し'],
      pkgs_quantity: @hostinfo[:pkgs] ? @hostinfo[:pkgs].length : 0,
      apps: @hostinfo[:apps] ? @hostinfo[:apps].map{|app| "#{app.name} #{app.version}\n#{app.path}"} : ['なし'],
      lapps: @hostinfo[:lapps] ? @hostinfo[:lapps].map{|lapp| lapp.get_name } : ['なし'],
      memo: @host.memo,
      hostname: @host.hostname,
      ext_name: @host.ext_name,
      display_name: @host.display_name,
    }   
    if f_params[:history] then
      history = JSON.parse(f_params[:history])
      html_id = history[0].to_i
      host_id = history[1].to_s
      count = 0
      @vuls[:content_filtered].each do |vul|
        if vul[:html_id] === html_id then
          break
        else 
          count += 1;
        end
      end
      @vuls[:offset] = count
      @history = html_id
      session['delete_histories'] = [host_id, html_id]
    end 
    @vuls[:content] = @vuls[:content_filtered].slice(@vuls[:offset], @vuls[:per_page])
    @vuls[:content] = [] unless @vuls[:content]

    #ブラウザバックかどうか判断するための処理
    if !session['chkbok_ids'] then
      session['chkbok_ids'] = {}
    end
    if !session['chkbok_ids']['host'] then
      session['chkbok_ids']['host'] = {}
    end
    session['chkbok_ids']['host']['is_browser_back'] = false
  end

  def update_status
    f_params = params.fetch(:status, {}).permit(:htmlids, :prev_status, :new_status, :comment, :dump, :action)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    if !f_params[:htmlids] or f_params[:htmlids].size <= 1 then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    htmlids = nil
    begin
      htmlids = JSON.parse(f_params[:htmlids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !htmlids.respond_to?(:'[]') or !htmlids.respond_to?(:size) or htmlids.size == 0 then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin then
      check = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:sidfmd_host_id => @user_host_ids)
        .where(:html_id => htmlids)
        .count
      if check == 0 then
        dump[:notice] = 'アクセスを許可されておりません。'
        redirect_to dump and return
      end
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => @id)
          .where(:html_id => htmlids)
        if prev_status > 0 then
          issues = issues
            .where(:issues => { :status_id => prev_status })
        end
        issues.each do |issue|
          issue.with_lock do
            issue.change_status_with_journal(User.current.id, new_status, comment)
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = 'ステータスの変更に失敗しました。'
      redirect_to dump and return
    end

    action = f_params[:action]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['host'] then
        session['hostid_with_htmlid']['host'] = {}
      end
      session['hostid_with_htmlid']['host'][@id.to_s] = [htmlids[0].to_s, true]
      dump[:'form[is_update_one_row]'] = true
    end
    
    redirect_to dump
  end

  def update_due_dates
    f_params = params.fetch(:schedule, {}).permit(:htmlids, :force, :due_date, :comment, :dump, :action)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    htmlids = JSON.parse(f_params[:htmlids])
    if htmlids.size == 0 or !htmlids.respond_to?(:'[]') then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin then
      check = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:sidfmd_host_id => @user_host_ids)
        .where(:html_id => htmlids)
        .count
      if check == 0 then
        dump[:notice] = 'アクセスを許可されておりません。'
        redirect_to dump and return
      end
    end
    force = !f_params[:force].nil?
    due_date = f_params[:due_date]
    unless due_date then
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        log_exception(e)
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          log_exception(e)
        end
      end
      if temp then
        due_date = temp
      else
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => @id)
          .where(:html_id => htmlids)
          .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
        if !force then
          issues = issues
            .where(:issues => { :due_date => nil })
        end
        issues.each do |issue|
          issue.with_lock do
            issue.change_due_date_with_journal(User.current.id, due_date, comment)
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end
    action = f_params[:action]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['host'] then
        session['hostid_with_htmlid']['host'] = {}
      end
      session['hostid_with_htmlid']['host'][@id.to_s] = [htmlids[0].to_s, true]
      dump[:'form[is_update_one_row]'] = true
    end

    redirect_to dump
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

  def download
    pages_params = params.fetch(:download, {}).permit(:format, :idx, :order, :filters, :htmlids)
    format =  pages_params[:format]
    if ![ 'json', 'tsv', 'playbook' ].include?(format) then
      redirect_to :action => :index , :notice => '不正な入力です。' and return
    end
    @vuls = { :url => request.url }
    @vuls[:idx] = (pages_params[:idx] || 'sri').to_sym
    @vuls[:order] = (pages_params[:order] || 'title_asc').to_sym
    @vuls[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
    @vuls[:htmlids] = []
    begin
      @vuls[:htmlids] = JSON.parse(pages_params[:htmlids] || '[]')
    rescue JSON::ParserError => e
      logger.error(e)
    end
    check_filters
    @corresp_required = Sidfm.corresp_required
    create_all_content
    @vuls[:corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:warning][@vuls[:idx]] ? 1 : 0 }
    @vuls[:fixed_corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:fixed_warning][@vuls[:idx]] ? 1 : 0 }
    @vuls[:not_fixed_corresp_required_vul_size] =
      @vuls[:corresp_required_vul_size] - @vuls[:fixed_corresp_required_vul_size]
    @vuls[:all_vul_size] = @vuls[:all_content].size
    filter_content
    @hostinfo = create_host_info
   if format == 'json' || format == 'playbook' then
      json = {}
      temp_order = @vuls[:order].to_s.split(/_/)
      json[:order] = { :by => temp_order[0], :type => temp_order[1], }
      json[:filters] = @vuls[:filters]
      json[:idx] = @vuls[:idx]
      json[:host] = {
        :id => @host.id,
        :hostname => @host.hostname,
        :os => (@hostinfo[:os] and @hostinfo[:os][:osname]) ? @hostinfo[:os][:osname] : '---',
        :last_checked_at =>  @host.last_checked_at ? @host.last_checked_at.strftime('%Y/%m/%d %H:%M:%S') : ( @hostinfo[:queue] ? @hostinfo[:queue].created_at.strftime('%Y/%m/%d %H:%M:%S') : '未同期' ),
        :is_agent => @host.is_agent,
      }
      temp_s_cat = @host.system_category
      json[:host][:system_category] = temp_s_cat ? { :id => temp_s_cat.id, :name => temp_s_cat.name } : { :id => 0, :name => '---' }
      json[:host][:no_system_categories] = @host.no_system_categories.collect {|x| { :id => x.sidfmd_category_id, :name => x.sidfmd_category.name } }
      temp_pkgs = @hostinfo[:pkgs]
      json[:host][:packages] = temp_pkgs ? temp_pkgs.collect {|x| x.pkg } : []
      temp_apps = @hostinfo[:apps]
      json[:host][:products] = temp_apps ? temp_apps.collect {|x| { :name => x.name, :version => x.version, :path => x.path } } : []
      temp_lapps = @hostinfo[:lapps]
      json[:host][:manual_products] = temp_lapps ? temp_lapps.collect {|x| x.get_name } : []
      json[:host][:memo] = @host.memo
      json[:content] = []
      @vuls[:content_filtered].each do |vul|
        if @vuls[:htmlids].size != 0 and !@vuls[:htmlids].include?(vul[:html_id]) then
          next
        end
        item = {}
        if format == 'json' then 
          item[:html_id] = vul[:html_id] > 0 ? "#{vul[:html_id]}" : "KB#{-vul[:html_id]}"
          item[:issue_id] = vul[:issue_id]
          item[:title] = vul[:title]
          item[:short_description] = (vul[:short_description] ? vul[:short_description] : '')
          item[:warning] = vul[:warning][@vuls[:idx]]
          item[:status] = SidfmdIssue.status_id_to_s(vul[:status_id])
          item[:sri] = SidfmdIssueSri::RISKS[vul[:sri]]
          t_cvss = vul[:cvss]
          item[:cvss] = (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f))
          item[:area] = SidfmdArea.where(:area_id => vul[:softek][:area_id]).first.name
          item[:exploitable] = (t_cvss[:label] != :notreg ? vul[:exploitable] : nil)
          item[:elapsed_days] = (vul[:created_on][:elapsed_days] >= 0 ? vul[:created_on][:elapsed_days] : '')
          item[:created_on] = vul[:created_on][:display]
          item[:due_date] = (vul[:due_date] ? vul[:due_date].strftime('%Y/%m/%d') : '')
          item[:journals] = []
          vul[:journals].each do |journal|
            j_item = {}
            j_item[:id] = journal[:id]
            j_item[:created_on] = journal[:created_on] ? journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') : '---'
            j_item[:user] = { :id => journal[:user_id], :name => view_context.display_user_name(User.find(journal[:user_id])) }
            j_item[:notes] = Journal.find(journal[:id]).notes
            j_item[:details] = []
            JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
              jd_item = {}
              jd_item[:id] = detail[:id]
              case detail[:prop_key]
              when 'status_id' then
                jd_item[:action] = 'ステータス変更'
                jd_item[:from] = SidfmdIssue.status_id_to_s(detail[:old_value].to_i)
                jd_item[:to] = SidfmdIssue.status_id_to_s(detail[:value].to_i)
              when 'due_date' then
                jd_item[:action] = '対策予定日の変更'
                jd_item[:from] = SidfmdIssue.due_date_to_s(detail[:old_value])
                jd_item[:to] = SidfmdIssue.due_date_to_s(detail[:value])
              else
                next
              end
              j_item[:details] << jd_item
            end
            item[:journals] << j_item
          end
        end
        item[:affect_packages] = []
        item[:affect_products] = []
        issue = SidfmdIssue.find(vul[:sidfmd_issue_id])
        if issue then
          item[:affect_packages] = issue.affect_packages('')
          item[:affect_products] = issue.affect_products
        end
        json[:content] << item
      end
      
      if format == 'playbook' then
        a = GenerateAnsiblePlaybook.new(json.to_json)
        playbook = a.generate
        if playbook != 'error' && playbook != 'no_packages' then 
          send_data(playbook, :filename => "vmplaybook-#{@host.hostname}-yml-#{Time.current.strftime('%Y%m%d%H%M%S')}.yml", :disposition => 'attachment')
          return 
        end
      else
        send_data(json.to_json, :filename => "vmlists-#{@host.hostname}-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json", :disposition => 'attachment')
        return
      end

    else
      header = [ 'HTML_ID', '要対応', '対応状況', 'SRI', 'CVSS', '影響範囲', '攻撃コード', '経過日数', '検出日', '脆弱性タイトル', '概要', '履歴', '対策予定日' ]
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @vuls[:content_filtered].each do |vul|
          if @vuls[:htmlids].size != 0 and !@vuls[:htmlids].include?(vul[:html_id]) then
            next
          end
          item.clear
          item << (vul[:html_id] > 0 ? "#{vul[:html_id]}" : "KB#{-vul[:html_id]}").encode('UTF-16LE')
          item << vul[:warning][@vuls[:idx]].to_s.encode('UTF-16LE')
          item << SidfmdIssue.status_id_to_s(vul[:status_id]).encode('UTF-16LE')
          item << SidfmdIssueSri::RISKS[vul[:sri]].encode('UTF-16LE')
          t_cvss = vul[:cvss]
          item << (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f)).encode('UTF-16LE')
          item << SidfmdArea.where(:area_id => vul[:softek][:area_id]).first.name.encode('UTF-16LE')
          item << (t_cvss[:label] != :notreg ? vul[:exploitable].to_s : '').encode('UTF-16LE')
          item << (vul[:created_on][:elapsed_days] >= 0 ? vul[:created_on][:elapsed_days].to_s : '').encode('UTF-16LE')
          item << vul[:created_on][:display].encode('UTF-16LE')
          item << vul[:title].encode('UTF-16LE')
          item << (vul[:short_description] ? vul[:short_description] : '').encode('UTF-16LE')
          item << vul[:journals].size.to_s.encode('UTF-16LE')
          item << (vul[:due_date] ? vul[:due_date].strftime('%Y/%m/%d') : '').encode('UTF-16LE')
          o << item
        end
      end
      send_data(csv, :filename => "vmlists-#{@host.hostname}-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    end

    redirect_to :action => :index, :notice => 'ファイルの出力に失敗しました。'
    
  end

  def send_mail
    filtered_params = params.fetch(:form, {}).permit(:recipients, :subject, :message)
  begin
    mail = {}
    mail[:from] = User.current.mail
    email_with_name = []
    recipients = JSON.parse(filtered_params[:recipients])
    recipients.each do |key, val|
      email_name = %("#{val}" <#{key}>)
      email_with_name << email_name
    end
    #mail[:to] = email_with_name
    mail[:subject] = filtered_params[:subject]
    mail[:message] = filtered_params[:message]
    send_mail_num = SidfmdMailer::SEND_MAIL_NUM
    time_out_second = SidfmdMailer::SEND_MAIL_TIMEOUT_SECOND
    while (send_mail_targets = email_with_name.shift(send_mail_num)).size > 0 do
      threads = send_mail_targets.map{|fullname_and_address|
        Thread.new(mail.deep_dup, fullname_and_address){|cp_mail, data|
          ret = {}
          cp_mail[:to] = data
          begin
            SidfmdMailer.send_mail(cp_mail).deliver_now
          rescue Exception => e
            ret[:exception] = e
          end
          ret
        }
      }
      threads.each do |result|
        result = result.join(time_out_second)
        if result then
          result = result.value
          if result[:exception] then logger.error("failed send mail. #{result[:exception]}") end
        else
          logger.error("time-out of sending a mail")
        end
      end
    end
    #SidfmdMailer.send_mail(mail).deliver_now
  rescue Exception => e
    log_exception(e)
    redirect_to :action => :index, :notice => e and return
  end
    redirect_to :action => :index and return
  end

  def get_journal
    filtered_params = params.fetch(:data, {}).permit(:html_id)
    data = []
    html_id = filtered_params[:html_id] 
    issue_id = SidfmdIssue.find_by(:html_id => html_id).id
    journal = Journal.where(:journalized_id => issue_id).map{|j| [j.user_id, j.notes, j.created_on]}
    users = Hash[User.where(:id => journal.map{|j| j[0]}).map{|user| [user.id, user.login]}]
    journal.each  do |j|
      j.push(users[j[0]])
      data.push(j)
    end
    render :layout => false, :json => data.to_json
  end
 
end
