# -*- coding: utf-8 -*-
require 'date'
require 'sidfm_controller_support'

class SidfmVulnerabilitiesController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private

  include SidfmControllerSupport
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license, :except => [ :threshold_setting, :threshold_update ]
  before_action :check_roles

  before_action :get_view_state
  after_action :put_view_state

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    if [ 'update_status', 'update_due_date' ].include?(action_name) and !@sidfm_roles.include?(:use) then
      redirect_to sidfm_vulnerabilities_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    if [ 'threshold_update' ].include?(action_name) and !@sidfm_roles.include?(:admin) then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).order(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id] }
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
                     end
    @cache_prefix = @role_admin ? '/vuls/admin/0' : "/vuls/user/#{@user_category_ids.join(",")}"
  end

  # 対象となるコンテンツ全てを含む配列を返す。
  def create_all_content(category_id = 0)
    @vuls[:all_content] = SidfmdIssue.create_all_content(category_id: @vuls[:category_id], role_admin: @role_admin, corresp_required: @corresp_required, user_host_ids: @user_host_ids, user_category_ids: @user_category_ids)
    @enable_host_ids = SidfmdHost.enable_hosts.collect {|x| x.id }
    @category_host_ids =
      if @vuls[:category_id] > 0 then
        SidfmdCategoriesHost
          .where(:sidfmd_category_id => @vuls[:category_id])
          .select(:sidfmd_host_id).distinct.collect {|x| x.sidfmd_host_id }
      else
        SidfmdCategoriesHost.select(:sidfmd_host_id).distinct.collect {|x| x.sidfmd_host_id }
      end
  end

  TIMELINE_SPAN = 6
  TIMELINE_SPAN_UNIT = :month

  # タイムラインのデータを返す
  def get_timeline_data
    ret = {}
    ret[:now] = Time.current
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @vuls[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical', 5 => 'fixed',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大', 5 => '済',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1, :fixed ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
        :fixed => '済',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end

    flg_fixed = (SidfmdVulTimeline.column_names.include?('sri_fixed') && SidfmdVulTimeline.column_names.include?('cvss_fixed'))

    ret[:timescale].each do |timescale|
      stat = SidfmdVulTimeline.where('dt >= ?', timescale[:time])
      unless timescale[:is_first] then
        stat = stat.where('dt <= ?', ret[:timescale][timescale[:step] - 1][:time])
      end
      stat = stat.order(:dt => :desc).limit(1).first
      if stat then
        ret[:status_counts][SidfmdIssue::STATUS_NOTCONFIRMED][timescale[:step]] = stat.status_notconfirmed
        ret[:status_counts][SidfmdIssue::STATUS_ONGOING][timescale[:step]] = stat.status_ongoing
        ret[:status_counts][SidfmdIssue::STATUS_FIXED][timescale[:step]] = stat.status_fixed
        ret[:status_counts][SidfmdIssue::STATUS_NOTAFFECT][timescale[:step]] = stat.status_notaffect
        case @vuls[:idx]
        when :sri then
          ret[:risk_counts][5][timescale[:step]] = stat.sri_fixed if flg_fixed
          ret[:risk_counts][4][timescale[:step]] = stat.sri_critical
          ret[:risk_counts][3][timescale[:step]] = stat.sri_high
          ret[:risk_counts][2][timescale[:step]] = stat.sri_medium
          ret[:risk_counts][1][timescale[:step]] = stat.sri_low
          ret[:risk_counts][0][timescale[:step]] = stat.sri_none
        when :cvss then
          ret[:risk_counts][:fixed][timescale[:step]] = stat.cvss_fixed if flg_fixed
          ret[:risk_counts][:d4][timescale[:step]] = stat.cvss_d4
          ret[:risk_counts][:d3][timescale[:step]] = stat.cvss_d3
          ret[:risk_counts][:d2][timescale[:step]] = stat.cvss_d2
          ret[:risk_counts][:d1][timescale[:step]] = stat.cvss_d1
          ret[:risk_counts][:notreg][timescale[:step]] = stat.cvss_notreg
        end
      else
        ret[:status_counts][SidfmdIssue::STATUS_NOTCONFIRMED][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_ONGOING][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_FIXED][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_NOTAFFECT][timescale[:step]] = 0
        case @vuls[:idx]
        when :sri then
          ret[:risk_syms].each_key do |sri|
            ret[:risk_counts][sri][timescale[:step]] = 0
          end
        when :cvss then
          ret[:risk_labels].each do |cvss|
            ret[:risk_counts][cvss][timescale[:step]] = 0
          end
        end
      end
    end

    ret
  end

  # パラメータ解析
  def parse_params
    pages_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :category_id, :grp, :idx, :filters, :checked, :all_checked, :anchor)

    @vuls = { :url => request.url }

    @vuls[:category_id] = (pages_params[:category_id] || 0).to_i
    @vuls[:category_name] = '全ての脆弱性'
    if @vuls[:category_id] > 0 then
      cat = SidfmdCategory.find(@vuls[:category_id])
      if cat then
        if @role_admin then
          @vuls[:category_name] = cat.name
        else
          if cat.is_system then
            @vuls[:category_name] = cat.name
          elsif @user_category_ids.include?(@vuls[:category_id]) then
            @vuls[:category_name] = cat.name
          else
            @vuls[:category_id] = 0
          end
        end
      else
        @vuls[:category_id] = 0
      end
    else
      @vuls[:category_id] = 0
    end

    @vuls[:grp] = (pages_params[:grp] || 'count').to_sym
    @vuls[:idx] = (pages_params[:idx] || @view_state[:idx].to_s).to_sym
    @view_state[:idx] = @vuls[:idx]
    @vuls[:per_page] = (pages_params[:per_page] || 10).to_i
    @vuls[:offset] = (pages_params[:offset] || 0).to_i
    @vuls[:order] = pages_params[:order] || 'title_asc'
    @vuls[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
    @vuls[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @vuls[:all_checked] = pages_params[:all_checked] || false
  end

  def create_content
    @corresp_required = Sidfm.corresp_required

    create_all_content(@vuls[:category_id])

    check_filters

    @vuls[:content_filtered] = apply_filters(@vuls[:all_content])
    @vuls[:total_size] = @vuls[:content_filtered].size
    @vuls[:content] = @vuls[:content_filtered].slice(@vuls[:offset], @vuls[:per_page])
    @vuls[:content] = [] unless @vuls[:content]


    @vuls[:all_vul_size] = @vuls[:all_content].size
    @vuls[:corresp_required_vul_size] = @vuls[:all_vul_size]
    @vuls[:fixed_corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:fixed_warning][@vuls[:idx]] ? 1 : 0 }
    @vuls[:not_fixed_corresp_required_vul_size] =
      @vuls[:all_content].sum {|x| x[:not_fixed_warning][@vuls[:idx]] ? 1 : 0 }

    @timeline_data = (@vuls[:grp] == :distribution) ? get_timeline_data : []
  end
  
  def filter_content
    @vuls[:distribution_status] = { :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0 }
    @vuls[:distribution_risk] =
      case @vuls[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end
    @vuls[:all_content].each do |i|
      case i[:min_status_id]
      when SidfmdIssue::STATUS_NOTCONFIRMED then @vuls[:distribution_status][:notconfirmed] += 1
      when SidfmdIssue::STATUS_ONGOING then @vuls[:distribution_status][:ongoing] += 1
      when SidfmdIssue::STATUS_FIXED then @vuls[:distribution_status][:fixed] += 1
      when SidfmdIssue::STATUS_NOTAFFECT then @vuls[:distribution_status][:notaffect] += 1
      else @vuls[:distribution_status][:notaffect] += 1
      end
      if SidfmdIssue::STATUS_UNFINISHED.include?(i[:min_status_id]) then
        case @vuls[:idx]
        when :sri then
          case i[:max_sri]
          when SidfmdIssueSri::VALUES[0] then
            @vuls[:distribution_risk][:critical] += 1
          when SidfmdIssueSri::VALUES[1] then
            @vuls[:distribution_risk][:high] += 1
          when SidfmdIssueSri::VALUES[2] then
            @vuls[:distribution_risk][:medium] += 1
          when SidfmdIssueSri::VALUES[3] then
            @vuls[:distribution_risk][:low] += 1
          when SidfmdIssueSri::VALUES[4] then
            @vuls[:distribution_risk][:none] += 1
          end
        when :cvss then
          case i[:cvss][:label]
          when :notreg then
            @vuls[:distribution_risk][:notreg] += 1
          when :d4 then
            @vuls[:distribution_risk][:d4] += 1
          when :d3 then
            @vuls[:distribution_risk][:d3] += 1
          when :d2 then
            @vuls[:distribution_risk][:d2] += 1
          when :d1 then
            @vuls[:distribution_risk][:d1] += 1
          end
        end
      else
        @vuls[:distribution_risk][:fixed] += 1
      end
    end

    @vuls[:dist_cat_vuls] = get_dist_cat_vuls(@vuls[:category_id])
    @vuls[:dist_os_vuls], @vuls[:dist_os_vuls_size] = get_dist_os_vuls(@vuls[:category_id])
    @vuls[:statuses] = SidfmdIssue.statuses

    @system_categories = SidfmdCategory.system
    @not_system_categories = SidfmdCategory.not_system
    if !@role_admin then
      @not_system_categories = @not_system_categories
        .where(:id =>
               SidfmdCategoriesUser
                 .select(:sidfmd_category_id)
                 .where(:user_id => User.current.id)
                 .distinct
                 .collect {|x| x[:sidfmd_category_id] })
    end

    @statuses = IssueStatus.all.order(:id)
  end

  FILTER_KEYS = {}

  # フィルター情報の初期化
  def init_filter_keys
    @due_dates = @vuls[:all_content].inject([]) {|r, x|
      r += x[:due_dates]
    }.uniq.collect {|x| x.nil? ? '' : x }.sort

    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:warning] = { :values => [ 'warning', 'others', 'all' ].map(&:freeze).freeze, :default => 'all'.freeze }
      FILTER_KEYS[:status] = {
        :values => (SidfmdIssue::STATUS_ALL.collect {|x| x.to_s } +
                    [ SidfmdIssue::STATUS_UNFINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ SidfmdIssue::STATUS_FINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ '0' ]).map(&:freeze).freeze, :default => '0'.freeze }
      FILTER_KEYS[:sri] = { :values => (SidfmdIssueSri::VALUES.collect {|x| x.to_s } + [ '' ]).map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:cvss] = { :values => [ '', 'notreg', 'd1', 'd2', 'd3', 'd4' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:area] = { :values => (SidfmdArea::VALUES.collect {|x| x.to_s } + [ '' ]).map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:exploitable] = { :values => [ 'true', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:due_date] = { :values => nil, :default => '::' }
      FILTER_KEYS[:keyword] = { :values => nil, :default => '' }
      FILTER_KEYS[:date] = { :values => nil, :default => '::' }
    end
  end

  # フィルタのチェックと正規化
  def check_filters
    init_filter_keys
    f_keys = FILTER_KEYS.keys
    for i in 0..f_keys.size - 1 do
      if @vuls[:filters].has_key?(f_keys[i]) then
        if FILTER_KEYS[f_keys[i]][:values] && !FILTER_KEYS[f_keys[i]][:values].include?(@vuls[:filters][f_keys[i]]) then
          @vuls[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
        end
      else
        @vuls[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
      end
    end
  end

  DAY_SECONDS = 24 * 60 * 60

  # フィルタの適用
  def apply_filters(all_content)
    ret = all_content.find_all {|item|
      # カテゴリ
      if @vuls[:category_id] > 0 then
        next false unless (item[:categories].collect {|x| x[:id] }).include?(@vuls[:category_id])
      end
      # 要注意
      if @vuls[:filters][:warning] == 'warning' then
        next false unless item[:warning][@vuls[:idx]]
      elsif @vuls[:filters][:warning] == 'others' then
        next false if item[:warning][@vuls[:idx]]
      end
      sri_cvss_status_ids = if @vuls[:filters][:warning] == 'warning' then
                              item[:sri_cvss_status_ids].find_all {|x| case @vuls[:idx] when :sri then x[:sri] >= @corresp_required[:sri]; when :cvss then x[:cvss] ? x[:cvss] >= @corresp_required[:cvss] : true; end }
                            else
                              item[:sri_cvss_status_ids]
                            end
      # 対応状況(new)
      # min_status_idで判断
      if @vuls[:filters][:status] != '0' then
        if /\A(\d):(\d)\z/ =~ @vuls[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (item[:min_status_id] == status1 or item[:min_status_id] == status2)
        else
          next false unless item[:min_status_id] == @vuls[:filters][:status].to_i
        end
      end

      # 対応状況(old)
      # sri_cvss_status_idsで判断
=begin
      if @vuls[:filters][:status] != '0' then
        if /\A(\d):(\d)\z/ =~ @vuls[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (sri_cvss_status_ids.find {|x| x[:status_id] == status1} or sri_cvss_status_ids.find {|x| x[:status_id] == status2 })
        else
          next false unless
            (sri_cvss_status_ids.size == 0 and @vuls[:filters][:status].to_i == SidfmdIssue::STATUS_NOTCONFIRMED) or
            (sri_cvss_status_ids.find {|x| x[:status_id] == @vuls[:filters][:status].to_i })
        end
      end
=end
      # 最大SRI
      if @vuls[:filters][:sri] != '' then
        next false unless item[:max_sri] == @vuls[:filters][:sri].to_i
      end
      # 最大CVSS
      if @vuls[:filters][:cvss] != '' then
        next false unless item[:cvss][:label] == @vuls[:filters][:cvss].to_sym
      end
      # SofTek影響範囲
      if @vuls[:filters][:area] != '' then
        next false unless (item[:html_id] > 0 and @vuls[:filters][:area].to_i == item[:softek][:area_id])
      end
      # 攻撃コード
      if @vuls[:filters][:exploitable] != '' then
        next false unless item[:exploitable]
      end
      # 対策予定日設定済
      if @vuls[:filters][:due_date] != '::' then
        if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @vuls[:filters][:due_date] then
          notset = ('notset' == Regexp.last_match(1))
          begindate = Regexp.last_match(2)
          enddate = Regexp.last_match(3)
          if notset then
            next false unless item[:due_dates].include?(nil)
          else
            begin
              begindate = Date.parse(begindate)
            rescue Exception => e
              log_exception(e)
              begindate = nil
            end
            begin
              enddate = Date.parse(enddate)
            rescue Exception => e
              log_exception(e)
              enddate = nil
            end
            due_date = nil
            if begindate and enddate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  log_exception(e)
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date >= begindate and d_due_date <= enddate then
                  flag = true
                  break
                end
              end
              next false unless flag
            elsif begindate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  log_exception(e)
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date >= begindate then
                  flag = true
                  break
                end
              end
              next false unless flag
            elsif enddate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  log_exception(e)
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date <= enddate then
                  flag = true
                  break
                end
              end
              next false unless flag
            end
          end
        end
      end
      # キーワード検索
      if @vuls[:filters][:keyword].size > 0 then
        if /\A\s*(\d+)\s*\z/ =~ @vuls[:filters][:keyword] then
          html_id = Regexp.last_match(1).to_i
          next false unless (Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE) =~ item[:title] or html_id == item[:html_id])
        elsif /\A\s*KB(\d+)\s*\z/i =~ @vuls[:filters][:keyword] then
          kbid = -Regexp.last_match(1).to_i
          next false unless (Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE) =~ item[:title] or kbid == item[:html_id])
        else
          next false unless Regexp.new('.*' + Regexp.escape(@vuls[:filters][:keyword]) + '.*', Regexp::IGNORECASE) =~ item[:title]
        end
      end
      # 日付指定検索
      if @vuls[:filters][:date] != '::' then
        if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @vuls[:filters][:date] then
          orig_flg_rev = Regexp.last_match(1)
          orig_begin_date = Regexp.last_match(2)
          orig_end_date = Regexp.last_match(3)
          if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
            @vuls[:filters][:date] = '::'
            next false
          end
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if begin_date <= item[:min_created_on][:value] and end_date >= item[:min_created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false if begin_date <= item[:min_created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if end_date >= item[:min_created_on][:value]
            else
              # last resort
              @vuls[:filters][:date] = '::'
              next false
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless begin_date <= item[:min_created_on][:value] and end_date >= item[:min_created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false unless begin_date <= item[:min_created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless end_date >= item[:min_created_on][:value]
            else
              # last resort
              @vuls[:filters][:date] = '::'
              next false
            end
          end
        else
          @vuls[:filters][:date] = '::'
          next false
        end
      end
      true
    }
    #ret.sort! {|x, y|
    #  SidfmdIssue::SORT_PRIORITIES[x[:min_status_id]] <=> SidfmdIssue::SORT_PRIORITIES[y[:min_status_id]] ||
    #  y[:min_created_on][:value] <=> x[:min_created_on][:value]
    #}
    case @vuls[:order]
    when 'warning_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:warning][@vuls[:idx]] ? 500 : 1000 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'warning_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:warning][@vuls[:idx]] ? 500 : 1000 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'statuses_asc' then
      temp_ev = proc {|x|
        m = x[:status_ids].min
        if m then
          [ m, 500 ]
        else
          [ '', 0 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'statuses_desc' then
      temp_ev = proc {|x|
        m = x[:status_ids].min
        if m then
          [ m, 500 ]
        else
          [ '', 0 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'maxsri_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:max_sri] ? x[:max_sri] : -100 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'maxsri_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:max_sri] ? x[:max_sri] : -100 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'cvss_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:cvss][:score] ? x[:cvss][:score] : -10 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'cvss_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:cvss][:score] ? x[:cvss][:score] : -10 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'area_asc' then
      temp_ev = proc {|x|
        [
         x[:title],
         x[:min_created_on][:display],
         x[:html_id] > 0 ? x[:softek][:area_id] : -100,
        ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'area_desc' then
      temp_ev = proc {|x|
        [
         x[:title],
         x[:min_created_on][:display],
         x[:html_id] > 0 ? x[:softek][:area_id] : -100,
        ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'exploitable_asc' then
      temp_ev = proc {|x|
        [ x[:title],
          x[:min_created_on][:display],
          x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 1000) : 1500) : 1250 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'exploitable_desc' then
      temp_ev = proc {|x|
        [ x[:title],
          x[:min_created_on][:display],
          x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 1000) : 1500) : 1250 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'elapseddays_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:min_created_on][:elapsed_days] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'elapseddays_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:min_created_on][:display], x[:min_created_on][:elapsed_days] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[2] <=> xx[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'title_asc' then
      temp_ev = proc {|x|
        [ x[:title], x[:html_id] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'title_desc' then
      temp_ev = proc {|x|
        [ x[:title], x[:html_id] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'hostnames_asc' then
      ret.sort! {|x, y|
        (x[:hosts].first[:ext_name] ? x[:hosts].first[:ext_name] : x[:hosts].first[:hostname]) <=> (y[:hosts].first[:ext_name] ? y[:hosts].first[:ext_name] : y[:hosts].first[:hostname])
        #x[:hosts].first[:hostname] <=> y[:hosts].first[:hostname]
      }
    when 'hostnames_desc' then
      ret.sort! {|y, x|
        (x[:hosts].first[:ext_name] ? x[:hosts].first[:ext_name] : x[:hosts].first[:hostname]) <=> (y[:hosts].first[:ext_name] ? y[:hosts].first[:ext_name] : y[:hosts].first[:hostname])
        #x[:hosts].first[:hostname] <=> y[:hosts].first[:hostname]
      }
    when 'categories_asc' then
      temp_ev = proc {|x|
        scat = x[:categories]  ? x[:categories].find {|y| y[:is_system] } : nil
        if scat then
          [ x[:title], x[:min_created_on][:display], scat[:id] ]
        else
          if x[:categories].size > 0 then
            [ x[:title], x[:min_created_on][:display], x[:categories][0][:id] ]
          else
            [ x[:title], x[:min_created_on][:display], 99999 ]
          end
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'categories_desc' then
      temp_ev = proc {|x|
        scat = x[:categories]  ? x[:categories].find {|y| y[:is_system] } : nil
        if scat then
          [ x[:title], x[:min_created_on][:display], scat[:id] ]
        else
          if x[:categories].size > 0 then
            [ x[:title], x[:min_created_on][:display], x[:categories][0][:id] ]
          else
            [ x[:title], x[:min_created_on][:display], 99999 ]
          end
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = yy[0] <=> xx[0]
          end
        end
        r
      }
    when 'duedates_asc' then
      temp_ev = proc {|x|
        p = view_context.display_due_dates_percent_value(x[:due_dates])
        [ x[:title], x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = yy[2] <=> xx[2]
          if r == 0 then
            r = yy[1] <=> xx[1]
            if r == 0 then
              r = yy[0] <=> xx[0]
            end
          end
        end
        r
      }
    when 'duedates_desc' then
      temp_ev = proc {|x|
        p = view_context.display_due_dates_percent_value(x[:due_dates])
        [ x[:title], x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = yy[2] <=> xx[2]
          if r == 0 then
            r = yy[1] <=> xx[1]
            if r == 0 then
              r = yy[0] <=> xx[0]
            end
          end
        end
        r
      }
    when 'finished_asc' then
      temp_ev = proc {|x|
        p = x[:status_ids].size > 0 ? view_context.display_finished_percent_value(x[:status_ids]) : nil
        [ x[:title], x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = yy[2] <=> xx[2]
          if r == 0 then
            r = yy[1] <=> xx[1]
            if r == 0 then
              r = yy[0] <=> xx[0]
            end
          end
        end
        r
      }
    when 'finished_desc' then
      temp_ev = proc {|x|
        p = x[:status_ids].size > 0 ? view_context.display_finished_percent_value(x[:status_ids]) : nil
        [ x[:title], x[:min_created_on][:display], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[3] <=> yy[3]
        if r == 0 then
          r = yy[2] <=> xx[2]
          if r == 0 then
            r = yy[1] <=> xx[1]
            if r == 0 then
              r = yy[0] <=> xx[0]
            end
          end
        end
        r
      }
    end
    ret
  end

  def get_dist_cat_vuls(category_id = 0)
    ret = {}
    scats = [ :invalid, :public_server, :private_server, :mobile, :client, :shared_server, :virtual_host_server, :public_network, :private_network ]
    scats.each do |s|
      ret[s] = 0
    end
    hosts =
      SidfmdHost
      .enable_hosts
    if category_id > 0 then
      hosts =
        hosts
        .where(:id =>
               SidfmdCategoriesHost
                 .where(:sidfmd_category_id => category_id)
                 .select(:sidfmd_host_id)
                 .distinct)
    else
      hosts = hosts.categorized
    end
    hosts.each do |host|
      ret[scats[host.system_category.id]] += 1
    end
    ret
  end

  def get_dist_os_vuls(category_id = 0)
    ret = []
    ret_hash = {}
    hosts =
      SidfmdHost
      .enable_hosts
    if category_id > 0 then
      hosts =
        hosts
        .where(:id =>
               SidfmdCategoriesHost
                 .where(:sidfmd_category_id => category_id)
                 .select(:sidfmd_host_id)
                 .distinct)
    else
      hosts = hosts.categorized
    end
    hosts.each do |host|
      dist = host.get_os
      os = '（不明）'
      if dist and dist[:osname] and dist[:osname].size > 0 then
        os = dist[:osname]
      end
      if ret_hash.has_key?(os) then
        ret_hash[os] += 1
      else
        ret_hash[os] = 1
      end
    end
    ret_hash.keys.each do |k|
      ret << { :name => k, :value => ret_hash[k] }
    end
    ret_size = ret.sum {|x| x[:value] }
    temp_ev = proc {|x| [ (x[:name] == '（不明）' ? 1000 : 100), x[:value], x[:name] ]}
    ret.sort! {|x, y|
      xx = temp_ev.call(x)
      yy = temp_ev.call(y)
      r = xx[0] <=> yy[0]
      if r == 0 then
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[2] <=> yy[2]
        end
      end
      r
    }
    [ ret, ret_size ]
  end

  public

  def index
    parse_params
  end

  def index_r
    parse_params
    create_content
    filter_content
    @self_user_categories = SidfmdCategory.where(:id => (SidfmdCategoriesUser.where(:user_id => User.current).map(&:sidfmd_category_id))).map{|c| c.is_system ? Hash["system_c", c.name] : Hash["custom_c", c.name]}
    @filter_inherit = {
      :'pages[idx]' => @vuls[:idx],
      :'pages[category_id]' => @vuls[:category_id],
      :'pages[filters]' => @vuls[:filters].reject {|k, v|
        [ :keyword ].include?(k)
      }.to_json,
    }

    #ブラウザバックかどうか判断するための処理
    if !session['chkbok_ids'] then
      session['chkbok_ids'] = {}
    end
    if !session['chkbok_ids']['vuls'] then
      session['chkbok_ids']['vuls'] = {}
    end
    session['chkbok_ids']['vuls']['is_browser_back'] = false
  end

  def threshold_setting
    @sri_risks = [ 4, 3, 2 , 1 ]
    @corresp_required = Sidfm.corresp_required
  end

  def threshold_update
    filtered_params = params.fetch(:params, {}).permit(:sri, :cvss)
    if filtered_params[:sri] then
      Sidfm.corresp_required_sri = filtered_params[:sri].to_i
    end
    if filtered_params[:cvss] then
      Sidfm.corresp_required_cvss = filtered_params[:cvss].to_f
    end
    Rails.cache.clear
    redirect_to :action => :threshold_setting, :notice => '設定しました。'
  end

  def update_status
    f_params = params.fetch(:status, {}).permit(:htmlids, :prev_status, :new_status, :comment, :dump)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    htmlids = nil
    begin
      htmlids = JSON.parse(f_params[:htmlids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !htmlids.respond_to?(:'[]') or !htmlids.respond_to?(:size) or htmlids.size == 0 then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin then
      check = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:sidfmd_host_id => @user_host_ids)
        .where(:html_id => htmlids)
        .count
      if check == 0 then
        dump[:notice] = 'アクセスを許可されておりません。'
        redirect_to dump and return
      end
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then 
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    if new_status == 0 then
      dump[:notice] = '設定後ステータスが設定されなかったため、操作は行われておりません。'
      redirect_to dump and return
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:html_id => htmlids)
        if prev_status == 0 then
          issues = issues
            .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
        else
          issues = issues
            .where(:issues => { :status_id => prev_status })
        end
        issues.each do |issue|
          issue.with_lock do
            issue.change_status_with_journal(User.current.id, new_status, comment)
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = 'ステータスの変更に失敗しました。'
      redirect_to dump and return
    end

    redirect_to dump
  end

  def update_due_dates
    f_params = params.fetch(:schedule, {}).permit(:htmlids, :force, :due_date, :comment, :dump)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index_r
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index_r
    htmlids = nil
    begin
      htmlids = JSON.parse(f_params[:htmlids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    unless htmlids then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin then
      check = SidfmdIssue
        .enable_hosts
        .categorized
        .where(:sidfmd_host_id => @user_host_ids)
        .where(:html_id => htmlids)
        .count
      if check == 0 then
        dump[:notice] = 'アクセスを許可されておりません。'
        redirect_to dump and return
      end
    end
    force = !f_params[:force].nil?
    due_date = f_params[:due_date]
    unless due_date then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        log_exception(e)
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          log_exception(e)
        end
      end
      if temp then
        due_date = temp
      else
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .enable_hosts
          .categorized
          .joins(:issue)
          .where(:html_id => htmlids)
          .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
        if !force then
          issue = issues
            .where(:issues => { :due_date => nil })
        end
        issues.each do |issue|
          issue.with_lock do
            issue.change_due_date_with_journal(User.current.id, due_date, comment)
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end

    redirect_to dump
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

  def download
    pages_params = params.fetch(:download, {}).permit(:format, :category_id, :idx, :order, :filters, :htmlids)
    format = pages_params[:format]
    if ![ 'json', 'tsv' ].include?(format) then
      redirect_to :action => :index ,:notice => '不正な入力です。' and return
    end
    @vuls = { :url => request.url }
    @vuls[:category_id] = (pages_params[:category_id] || 0).to_i
    if @vuls[:category_id] > 0 then
      cat  = SidfmdCategory.find(@vuls[:category_id])
      if cat then
        @vuls[:category_name] = cat.name
      else
        @vuls[:category_id] = 0
        @vuls[:category_name] = '全ての脆弱性'
      end
    else
      @vuls[:category_name] = '全ての脆弱性'
    end
    @vuls[:idx] = (pages_params[:idx] || 'sri').to_sym
    @vuls[:htmlids] = []
    begin
      @vuls[:htmlids] = JSON.parse(pages_params[:htmlids] || '[]')
    rescue JSON::ParserError => e
      logger.error(s)
    end
    @corresp_required = Sidfm.corresp_required
    create_all_content(@vuls[:category_id])
    @vuls[:order] = 'title_asc'
    @vuls[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
    check_filters
    @vuls[:content_filtered] = apply_filters(@vuls[:all_content])

    case format
    when 'json' then
      json = {}
      temp_order = @vuls[:order].to_s.split(/_/)
      json[:order] = { :by => temp_order[0], :type => temp_order[1], }
      json[:filters] = @vuls[:filters]
      json[:category] = { :id => @vuls[:category_id], :name => @vuls[:category_name], }
      json[:idx] = @vuls[:idx]
      json[:content] = []

      @vuls[:content_filtered].each do |vul|
        if @vuls[:htmlids].size != 0 and !@vuls[:htmlids].include?(vul[:html_id]) then
          next
        end
        item = {}

        item[:html_id] = vul[:html_id] > 0 ? "#{vul[:html_id]}" : "KB#{-vul[:html_id]}"
        item[:title] = vul[:title]
        item[:short_description] = (vul[:short_description] ? vul[:short_description] : '')
        item[:warning] = vul[:warning][@vuls[:idx]]
        count_status_ids = SidfmdIssue.count_status_ids(vul[:status_ids])
        item[:statuses] = count_status_ids
          .sort {|x, y| x[:id] <=> y[:id] }
          .collect {|x| { :name => x[:name], :value => x[:value] } }
        item[:max_sri] = SidfmdIssueSri::RISKS[vul[:max_sri]]
        t_cvss = vul[:cvss]
        item[:max_cvss] = (t_cvss[:label] == :notreg ? '' :sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f))
        item[:max_area] = SidfmdArea.where(:area_id => vul[:softek][:area_id]).first.name
        item[:exploitable] = (t_cvss[:label] != :notreg ? vul[:exploitable] : nil)
        item[:min_elapsed_days] = (vul[:min_elapsed_created_on][:display] >= 0 ? vul[:min_elapsed_created_on][:display] : '')
        item[:min_created_on] = vul[:min_created_on][:display]
        item[:hosts] = vul[:hosts].collect {|x| x[:hostname] }
        item[:categories] = vul[:categories].collect { |x| x[:name] }
        item[:due_dates_percent] = (vul[:due_dates].size > 0 ? display_due_dates_percent(vul[:due_dates]) : '0.0%')

        item[:statuses_percent] = (vul[:status_ids].size > 0 ? display_finished_percent(vul[:status_ids]) : '0.0%')
        json[:content] << item
      end

      send_data(json.to_json, :filename => "vmlist-vulnerabilities-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json", :disposition => 'attachment')
      return
    when 'tsv' then
      header = [ 'HTML_ID', '要対応', '対応状況', '最大SRI', 'CVSS', '影響範囲', '攻撃コード', '経過日数', '検出日', '脆弱性タイトル', '概要', 'ホスト名', 'カテゴリ', '対策予定日設定済', '対処率' ]
      # TSV, UTF-16LE, BOMあり, 拡張子 .csv
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @vuls[:content_filtered].each do |vul|
          if @vuls[:htmlids].size != 0 and !@vuls[:htmlids].include?(vul[:html_id]) then
            next
          end
          item.clear
          item << (vul[:html_id] > 0 ? "#{vul[:html_id]}" : "KB#{-vul[:html_id]}").encode('UTF-16LE')
          item << vul[:warning][@vuls[:idx]].to_s.encode('UTF-16LE')
          count_status_ids = SidfmdIssue.count_status_ids(vul[:status_ids])
          item << count_status_ids.sort {|x, y| x[:id] <=> y[:id] }.collect {|x| "#{x[:name]}:#{x[:value]}" }.join(',').encode('UTF-16LE')
          item << (vul[:max_sri] ? SidfmdIssueSri::RISKS[vul[:max_sri]] : '').encode('UTF-16LE')
          t_cvss = vul[:cvss]
          item << (t_cvss[:label] == :notreg ? '' :sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f)).encode('UTF-16LE')
          item << SidfmdArea.where(:area_id => vul[:softek][:area_id]).first.name.encode('UTF-16LE')
          item << (t_cvss[:label] != :notreg ? vul[:exploitable].to_s : '').encode('UTF-16LE')
          item << (vul[:min_elapsed_created_on][:display] >= 0 ? vul[:min_elapsed_created_on][:display].to_s : '').encode('UTF-16LE')
          item << vul[:min_created_on][:display].encode('UTF-16LE')
          item << vul[:title].encode('UTF-16LE')
          item << (vul[:short_description] ? vul[:short_description] : '').encode('UTF-16LE')
          item << vul[:hosts].collect {|x| x[:hostname] }.join(',').encode('UTF-16LE')
          item << vul[:categories].collect { |x| x[:name] }.join(',').encode('UTF-16LE')
          item << (vul[:due_dates].size > 0 ? display_due_dates_percent(vul[:due_dates]) : '0.0%').encode('UTF-16LE')
          item << (vul[:status_ids].size > 0 ? display_finished_percent(vul[:status_ids]) : '0.0%').encode('UTF-16LE')
          o << item
        end
      end
      send_data(csv, :filename => "vmlist-vulnerabilities-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    end

    redirect_to :action => :index, :notice => 'ファイルの出力に失敗しました。'
  end
  
  def send_mail
    filtered_params = params.fetch(:form, {}).permit(:recipients, :subject, :message)
  begin
    mail = {}
    mail[:from] = User.current.mail
    email_with_name = []
    recipients = JSON.parse(filtered_params[:recipients])
    recipients.each do |key, val|
      email_name = %("#{val}" <#{key}>)
      email_with_name << email_name
    end
    #mail[:to] = email_with_name
    mail[:subject] = filtered_params[:subject]
    mail[:message] = filtered_params[:message]
    send_mail_num = SidfmdMailer::SEND_MAIL_NUM
    time_out_second = SidfmdMailer::SEND_MAIL_TIMEOUT_SECOND
    while (send_mail_targets = email_with_name.shift(send_mail_num)).size > 0 do
      threads = send_mail_targets.map{|fullname_and_address|
        Thread.new(mail.deep_dup, fullname_and_address){|cp_mail, data|
          ret = {}
          cp_mail[:to] = data
          begin
            SidfmdMailer.send_mail(cp_mail).deliver_now
          rescue Exception => e
            ret[:exception] = e
          end
          ret
        }
      }
      threads.each do |result|
        result = result.join(time_out_second)
        if result then
          result = result.value
          if result[:exception] then logger.error("failed send mail. #{result[:exception]}") end
        else
          logger.error("time-out of sending a mail")
        end
      end
    end
    #SidfmdMailer.send_mail(mail).deliver_now
  rescue Exception => e
    log_exception(e)
    redirect_to :action => :index, :notice => e and return
  end
    redirect_to :action => :index and return
  end

end
