# -*- coding: utf-8 -*-

class SidfmdRule < ActiveRecord::Base
  belongs_to :sidfmd_category

  has_one :sidfmd_rule_trigger, dependent: :destroy
  has_many :sidfmd_rule_conds, dependent: :destroy
  has_many :sidfmd_rule_actions, dependent: :destroy

  scope :enables, -> { where(:flg_enable => true) }
  scope :filter_category, -> (category_id) { where(:sidfmd_category_id => category_id) }

  #ルールの登録数上限
  MAXIMUM_LIMIT = 100

  #キーワードの変換時とhtmlのタイトル表示に使用する
  PAGE_TITLE = "ルール設定"

  #triggerの一覧。パラメータに使用する。
  RULE_TRIGGER_NAMES = {
      'periodic_hour' => "毎日実行-時間指定",
      'periodic_interval'=> "定期実行-間隔指定",
      'vulsync' => "脆弱性同期時 実行",
  }

  #condの一覧。htmlの要素やパラメータに使用する。
  RULE_COND_NAMES = {
      'issue-update' => '脆弱性の更新',
      'cvss-exploitable' => '攻撃コード指定',
      'sri-above' => 'SRI指定値以上',
      'sri-below' => 'SRI指定値以下',
      'sri-warning' => 'SRI値要対応',
      'cvss-above' => 'CVSS指定値以上',
      'cvss-below' => 'CVSS指定値以下',
      'cvss-warning' => 'CVSS値要対応',
      'cvss-notreg' => 'CVSS値未定義',
      'status-match' => 'ステータスの一致',
      'product-match' => 'プロダクト限定',
      'product-unmatch' => 'プロダクト除外',
      'due-date-expired' => '対策予定日超過',
      'due-date-approach' => '対策予定日接近',
      'history-create' => '履歴追加',
      'category-match' => 'カスタムカテゴリ限定',
      'category-unmatch' => 'カスタムカテゴリ除外',
      'host-match' => 'ホスト限定',
      'host-unmatch' => 'ホスト除外',
  }

  #action一覧。htmlの要素やパラメータに使用する。
  RULE_ACTION_NAMES = {
      'send-mail-user' => 'メール送信(ユーザ)',
      'send-mail-category'=> 'メール送信(カテゴリ)',
      'status-change' => 'ステータス変更',
      'due-date-change' => '対策予定日設定',
  }

  # ログでのTrigger, Cond, Actionの名称
  LOG_TRIGGER = 'trigger'.freeze
  LOG_COND = 'cond'.freeze
  LOG_ACTION = 'action'.freeze
  LOG_KINDS = {
    LOG_TRIGGER => 'タイミング'.freeze,
    LOG_COND => '条件'.freeze,
    LOG_ACTION => '操作'.freeze,
  }.freeze

  # ログの種類
  LOG_INFO = 'info'.freeze
  LOG_WARN = 'warn'.freeze
  LOG_ERR = 'err'.freeze
  LOG_TYPES = {
    LOG_INFO => '情報'.freeze,
    LOG_WARN => '警告'.freeze,
    LOG_ERR => 'エラー'.freeze,
  }.freeze

  # Trigger, Cond, Actionのそれぞれのlogsに保存できるログの最大件数
  # 1つのルールには1つのTriggerと複数のCond, Actionがあるので、合計するとそれなりの
  # 量になる。
  LOG_MAXNUM = 30

  # ログの追加処理
  # Trigger, Cond, Actionで利用
  def self.log_append(logs, type, kind, message)
    ret = nil
    begin
      ret = JSON.parse(logs)
    rescue JSON::ParserError
      ret = []
    end
    ret << { 'type' => type, 'kind' => kind, 'message' => message, 'created_at' => Time.current.iso8601(3) }
    ret.compact!
    ret.sort! {|x, y|
      (y['created_at'] && x['created_at']) ?
      y['created_at'] <=> x['created_at'] :
      (y['created_at'] ? 1 : -1)
    }
    JSON.generate(ret[0, LOG_MAXNUM])
  end

  # 指定した現在の条件に適合するルールのIDの配列を返す。
  def self.match_rule_ids(vulsync: false, periodic: false)
    ret = []
    # 定期実行
    if periodic then
      # hourは、前回実行時刻から20時間以上経過及び
      # 現在時刻がhour時になっていることが条件。
      # 始めての実行時(sidfmd_rule_triggers.last_checked_at=nil)は、後者の条件のみ。
      # current_timestampは、sidfmd_rule_triggers.last_checked_atがwithout time zone
      # だったため、整合性を採るために細工したもの。
      current_timestamp = "'#{Time.current.localtime.strftime('%Y/%m/%d %H:%M:%S')}'::timestamp without time zone"
      SidfmdRuleTpHour.enables
        .where("
          sidfmd_rule_tp_hours.hour IS NOT NULL AND (
            (
              sidfmd_rule_triggers.last_checked_at IS NULL AND
              TO_CHAR(#{current_timestamp}, 'HH24')::integer = sidfmd_rule_tp_hours.hour
            ) OR (
              sidfmd_rule_triggers.last_checked_at IS NOT NULL AND
              (#{current_timestamp} - sidfmd_rule_triggers.last_checked_at) >= '20 hours'::interval AND
              TO_CHAR(#{current_timestamp}, 'HH24')::integer = sidfmd_rule_tp_hours.hour
            )
          )
      ").each do |tph|
        ret << tph.sidfmd_rule_trigger.sidfmd_rule_id
      end
      # intervalは前回実行時間から、interval時間以上経過していることが条件
      # 初めての実行時(sidfmd_rule_triggers.last_checked_atがnil)は、無条件で適合。
      SidfmdRuleTpInterval.enables
        .where("
           (
             sidfmd_rule_tp_intervals.unit IS NOT NULL AND
             sidfmd_rule_tp_intervals.value IS NOT NULL ) AND (
             sidfmd_rule_triggers.last_checked_at IS NULL OR
             (#{current_timestamp} - sidfmd_rule_triggers.last_checked_at) >=
              (sidfmd_rule_tp_intervals.value::text || ' ' || sidfmd_rule_tp_intervals.unit || 's')::interval
           )
      ").each do |tpi|
        ret << tpi.sidfmd_rule_trigger.sidfmd_rule_id
      end
    end
    # 脆弱性同期時
    if vulsync then
      # vulsyncは無条件で適合。
      SidfmdRuleTVulsync.enables.each do |i|
        ret << i.sidfmd_rule_trigger.sidfmd_rule_id
      end
    end
    ### FIXME: 有効なruleのtriggerについて、logsに成否の結果を記録。
    ret
  end

  # ルールのTrigger, Cond, Actionのログをまとめて時系列でソートし返す。
  # 以下のHashの配列を返す。
  # {
  #   'type'       LOG_TYPES.keysのいずれか。LOG_TYPESで日本語名称を取得して出力すること。
  #   'kind'       LOG_KINDS.keysのいずれか。LOG_KINDSで日本語名称を朱徳して出力すること。
  #   'message'    ログ内容
  #   'created_at' 記録日時
  # }
  def logs(maxlogs = 200)
    ret = []
    if self.sidfmd_rule_trigger then
      begin
        r = JSON.parse(self.sidfmd_rule_trigger.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_TRIGGER)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    self.sidfmd_rule_conds.each do |cond|
      begin
        r = JSON.parse(cond.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_COND)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    self.sidfmd_rule_actions.each do |action|
      begin
        r = JSON.parse(action.logs)
        r.each do |i|
          ret << i.merge('kind' => SidfmdRule::LOG_ACTION)
        end
      rescue JSON::ParserError => e
        logger.error(e.message)
        e.backtrace.each do |i|
          logger.error(i)
        end
      end
    end
    ret.compact!
    ret.sort! {|x, y|
      (y['created_at'] && x['created_at']) ?
      y['created_at'] <=> x['created_at'] :
      (y['created_at'] ? 1 : -1)
    }
    # 時刻のフォーマットの変更
    ret.each do |i|
      if i['created_at'] then
        i['created_at'].gsub!(/^(\d{4}-\d{2}-\d{2})T(\d{2}:\d{2}:\d{2}\.\d+)[+-]\d{2}:\d{2}$/, '\1 \2')
      end
    end
    ret[0, maxlogs].reverse
  end

end
